Package: rslurm
Type: Package
Title: Submit R Calculations to a 'Slurm' Cluster
Description: Functions that simplify submitting R scripts to a 'Slurm' 
    workload manager, in part by automating the division of embarrassingly
    parallel calculations across cluster nodes.
Acknowledgements: Development of this R package was supported by the
        National Socio-Environmental Synthesis Center (SESYNC) under
        funding received from the National Science Foundation grants
        DBI-1052875 and DBI-1639145.
Version: 0.6.2
Date: 2023-01-31
License: GPL-3
URL: https://www.earthdatascience.org/rslurm/
BugReports: https://github.com/earthlab/rslurm/issues
Authors@R: c(person('Philippe', 'Marchand', email = "marchand.philippe@gmail.com", role = 'aut'),
             person('Ian', 'Carroll', role = 'aut'),
             person('Mike', 'Smorul', role = 'aut'),
             person('Rachael', 'Blake', role = 'aut'),
             person('Quentin', 'Read', email = 'quentin.read@usda.gov', role = 'aut'),
             person('Dayne', 'Filer', role = 'ctb'),
             person('Ben', 'Fasoli', role = 'ctb'),
             person('Pol', 'van Rijn', role = 'ctb'),
             person('Sebastian', 'Schubert', role = 'ctb'),
             person('Rob', 'Gilmore', role = 'ctb'),
             person('Christopher', 'Barrington', role = 'ctb'),
             person('Se Jong', 'Cho', role = 'art'),
             person('Erick Verleye', email='erve3705@colorado.edu', role = 'cre')
             )
Depends: R (>= 3.5.0)
Imports: whisker (>= 0.3)
RoxygenNote: 7.2.3
Suggests: parallel, testthat, knitr, rmarkdown, markdown
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2023-02-23 19:14:46 UTC; verleyee
Author: Philippe Marchand [aut],
  Ian Carroll [aut],
  Mike Smorul [aut],
  Rachael Blake [aut],
  Quentin Read [aut],
  Dayne Filer [ctb],
  Ben Fasoli [ctb],
  Pol van Rijn [ctb],
  Sebastian Schubert [ctb],
  Rob Gilmore [ctb],
  Christopher Barrington [ctb],
  Se Jong Cho [art],
  Erick Verleye [cre]
Maintainer: Erick Verleye <erve3705@colorado.edu>
Repository: CRAN
Date/Publication: 2023-02-24 08:42:35 UTC
Built: R 4.6.0; ; 2026-04-25 22:03:48 UTC; unix

# Generated by roxygen2: do not edit by hand

export(cancel_slurm)
export(cleanup_files)
export(get_job_status)
export(get_slurm_out)
export(local_slurm_array)
export(print_job_status)
export(slurm_apply)
export(slurm_call)
export(slurm_job)
export(slurm_map)
importFrom(parallel,mclapply)
importFrom(utils,capture.output)
importFrom(utils,read.table)

<!-- README.md is generated from README.Rmd. Please edit that file -->

<img src="man/figures/logo_slurm.png" align="right" width=150px>

# rslurm: submit R code to a Slurm cluster

<!-- badges: start -->

[![rstudio mirror
downloads](https://cranlogs.r-pkg.org/badges/rslurm)](https://CRAN.R-project.org/package=rslurm)
[\[R build
status\]https://github.com/earthlab/rslurm/workflows/R-CMD-check/badge.svg)](https://github.com/earthlab/rslurm/actions)
[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![CRAN
status](https://www.r-pkg.org/badges/version/rslurm)](https://CRAN.R-project.org/package=rslurm)
[![DOI](https://zenodo.org/badge/37485241.svg)](https://zenodo.org/badge/latestdoi/37485241)
[![R-CMD-check](https://github.com/earthlab/rslurm/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/earthlab/rslurm/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

### About

Development of this R package was supported by the National
Socio-Environmental Synthesis Center (SESYNC) under funding received
from the National Science Foundation grants DBI-1052875 and DBI-1639145.

The package was developed by Philippe Marchand and Ian Carroll, with
Mike Smorul and Rachael Blake contributing. Erick Verleye is the current
maintainer.

As of January 2023 the University of Colorado Boulder’s Earth Lab
adopted rslurm and is responsible for all future maintenance.

### Installation

You can install the released version of
[rslurm](https://cran.r-project.org/package=rslurm) from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("rslurm")
```

And the development version from
[GitHub](https://github.com/earthlab/rslurm) with:

``` r
# install.packages("devtools")
devtools::install_github("earthlab/rslurm")
```

### Documentation

Package documentation is accessible from the R console through
`package?rslurm` and [online](https://www.earthdatascience.org/rslurm/).

### Example

Note that job submission is only possible on a system with access to a
Slurm workload manager (i.e. a system where the command line utilities
`squeue` or `sinfo` return information from a Slurm head node).

To illustrate a typical rslurm workflow, we use a simple function that
takes a mean and standard deviation as parameters, generates a million
normal deviates and returns the sample mean and standard deviation.

``` r
test_func <- function(par_mu, par_sd) {
    samp <- rnorm(10^6, par_mu, par_sd)
    c(s_mu = mean(samp), s_sd = sd(samp))
}
```

We then create a parameter data frame where each row is a parameter set
and each column matches an argument of the function.

``` r
pars <- data.frame(par_mu = 1:10,
                   par_sd = seq(0.1, 1, length.out = 10))
```

We can now pass that function and the parameters data frame to
`slurm_apply`, specifying the number of cluster nodes to use and the
number of CPUs per node.

``` r
library(rslurm)
sjob <- slurm_apply(test_func, pars, jobname = 'test_apply',
                    nodes = 2, cpus_per_node = 2, submit = FALSE)
```

The output of `slurm_apply` is a `slurm_job` object that stores a few
pieces of information (job name, job ID, and the number of nodes) needed
to retrieve the job’s output.

See [Get
started](https://www.earthdatascience.org/rslurm/articles/rslurm.html)
for more information.

library(base, quietly = TRUE)
library(methods, quietly = TRUE)
library(datasets, quietly = TRUE)
library(utils, quietly = TRUE)
library(grDevices, quietly = TRUE)
library(graphics, quietly = TRUE)
library(stats, quietly = TRUE)
library(rslurm, quietly = TRUE)


.rslurm_func <- readRDS('f.RDS')
.rslurm_params <- readRDS('params.RDS')
.rslurm_result <- do.call(.rslurm_func, .rslurm_params)
               
saveRDS(.rslurm_result, file = 'results_0.RDS')

#' Prints the status of a Slurm job and, if completed, its console/error output
#'
#' Prints the status of a Slurm job and, if completed, its console/error output.
#'
#' If the specified Slurm job is still in the queue or running, this function
#' prints its current status (as output by the Slurm \code{squeue} command).
#' The output displays one row by node currently running part of the job ("R" in
#' the "ST" column) and how long it has been running ("TIME"). One row indicates
#' the portions of the job still in queue ("PD" in the "ST" column), if any. 
#' 
#' If all portions of the job have completed or stopped, the function prints the 
#' console and error output, if any, generated by each node.
#' 
#' @param slr_job A \code{slurm_job} object.
#' 
#' @name print_job_status-deprecated
#' @usage print_job_status(slr_job)
#' @seealso \code{\link{rslurm-deprecated}}
#' @keywords internal
NULL

#' @rdname rslurm-deprecated
#' @section \code{print_job_status}:
#' For \code{print_job_status}, use \code{\link{get_job_status}}.
#'
#' @export
print_job_status <- function(slr_job) {
    .Deprecated("get_job_status")
    
    if (!(inherits(slr_job, "slurm_job"))) stop("input must be a slurm_job")  
    stat <- suppressWarnings(
        system(paste("squeue -n", slr_job$jobname), intern = TRUE))
    if (length(stat) > 1) {
        cat(paste(c("Job running or in queue. Status:", stat), collapse = "\n"))
    } else {
        cat("Job completed or stopped. Printing console output below if any.\n")
        tmpdir <- paste0("_rslurm_", slr_job$jobname)
        out_files <- file.path(tmpdir, 
                               paste0("slurm_", 0:(slr_job$nodes - 1), ".out"))
        for (outf in out_files) {
            cat(paste("\n----", outf, "----\n\n"))
            cat(paste(readLines(outf), collapse = "\n"))
        }
    }
}  

# Utility functions for rslurm package (not exported)


# Make jobname by cleaning user-provided name or (if NA) generate one
# from base::tempfile
make_jobname <- function(name) {
    if (is.na(name)) {
        tmpfile <- tempfile("_rslurm_", tmpdir=".")
        strsplit(tmpfile, '_rslurm_', TRUE)[[1]][[2]]
    } else {
        jobname <- gsub("[[:space:]]+", "_", name)
        gsub("[^0-9A-Za-z_]", "", jobname)
    }
}


# Format sbatch options into nested list for templates
format_option_list <- function(slurm_options) {
    if (length(slurm_options) == 0) {
        slurm_flags <- slurm_options
    } else {
        is_flag <- sapply(slurm_options, isTRUE)
        slurm_flags <- lapply(names(slurm_options[is_flag]), function(x) {
            list(name = x)
        })
        slurm_options <- slurm_options[!is_flag]
        slurm_options <- lapply(seq_along(slurm_options), function(i) {
            list(name = names(slurm_options)[i], value = slurm_options[[i]])
        })        
    }
    list(flags = slurm_flags, options = slurm_options)
}

# Submit job
submit_slurm_job <- function(tmpdir) {
    old_wd <- setwd(tmpdir)
    sys_out <- tryCatch({
            system("sbatch submit.sh", intern = TRUE)
        }, finally = setwd(old_wd))
    message(sys_out)
    sys_out <- strsplit(sys_out, " ")[[1]]
    jobid <- sys_out[length(sys_out)]
    return(jobid)
}

# Submit dummy job with a dependency via sbatch to block R process
wait_for_job <- function(slr_job) {
    queued <- system(
        paste('test -z "$(squeue -hn', slr_job$jobname, '2>/dev/null)"'),
        ignore.stderr = TRUE)
    if (queued) {
        block_cmd <- sprintf('sbatch --nodes=1 --output=/dev/null --time=00:01:00 --dependency=singleton --job-name=%s --wait --wrap="hostname"', slr_job$jobname)
        system(block_cmd, intern=TRUE)
    }
    return()
}

#' Parallel execution of a function over a list on the Slurm cluster
#'
#' Use \code{slurm_map} to compute function over a list
#' in parallel, spread across multiple nodes of a Slurm cluster,
#' with similar syntax to \code{lapply}.
#'
#' This function creates a temporary folder ("_rslurm_[jobname]") in the current
#' directory, holding .RData and .RDS data files, the R script to run and the Bash
#' submission script generated for the Slurm job.
#'
#' The set of input parameters is divided in equal chunks sent to each node, and
#' \code{f} is evaluated in parallel within each node using functions from the
#' \code{parallel} R package. The names of any other R objects (besides
#' \code{x}) that \code{f} needs to access should be included in
#' \code{global_objects} or passed as additional arguments through \code{...}.
#'
#' Use \code{slurm_options} to set any option recognized by \code{sbatch}, e.g.
#' \code{slurm_options = list(time = "1:00:00", share = TRUE)}.
#' See \url{http://slurm.schedmd.com/sbatch.html} for details on possible options.
#' Note that full names must be used (e.g. "time" rather than "t") and that flags
#' (such as "share") must be specified as TRUE. The "array", "job-name", "nodes", 
#' "cpus-per-task" and "output" options are already determined by 
#' \code{slurm_map} and should not be manually set.
#'
#' When processing the computation job, the Slurm cluster will output two types
#' of files in the temporary folder: those containing the return values of the
#' function for each subset of parameters ("results_[node_id].RDS") and those
#' containing any console or error output produced by R on each node
#' ("slurm_[node_id].out").
#'
#' If \code{submit = TRUE}, the job is sent to the cluster and a confirmation
#' message (or error) is output to the console. If \code{submit = FALSE},
#' a message indicates the location of the saved data and script files; the
#' job can be submitted manually by running the shell command
#' \code{sbatch submit.sh} from that directory.
#'
#' After sending the job to the Slurm cluster, \code{slurm_map} returns a
#' \code{slurm_job} object which can be used to cancel the job, get the job
#' status or output, and delete the temporary files associated with it. See
#' the description of the related functions for more details.
#'
#' @param x A list to apply \code{f} to. Each
#'   element of \code{x} corresponds to a separate function call.
#' @param f A function that accepts one element of \code{x} as its first argument,
#' and may return any type of R object.
#' @param ... Additional arguments to \code{f}. These arguments do not vary
#' with each call to \code{f}.
#' @param jobname The name of the Slurm job; if \code{NA}, it is assigned a
#'   random name of the form "slr####".
#' @param nodes The (maximum) number of cluster nodes to spread the calculation
#'   over. \code{slurm_map} automatically divides \code{x} in chunks of
#'   approximately equal size to send to each node. Less nodes are allocated if
#'   the parameter set is too small to use all CPUs on the requested nodes.
#' @param cpus_per_node The number of CPUs requested per node. This argument is
#'   mapped to the Slurm parameter \code{cpus-per-task}.
#' @param processes_per_node The number of logical CPUs to utilize per node,
#'   i.e. how many processes to run in parallel per node. This can exceed
#'   \code{cpus_per_node} for nodes which support hyperthreading. Defaults to
#'   \code{processes_per_node = cpus_per_node}.
#' @param preschedule_cores Corresponds to the \code{mc.preschedule} argument of 
#'   \code{parallel::mclapply}. Defaults to \code{TRUE}. If \code{TRUE}, the 
#'   elements of \code{x} are assigned to cores before computation. 
#'   If \code{FALSE}, each element of \code{x} is executed by the next available core.
#'   Setting \code{FALSE} may be faster if 
#'   different elements of \code{x} result in very variable completion time for
#'   jobs.
#' @param job_array_task_limit The maximum number of job array tasks to run at 
#'   the same time. Defaults to \code{NULL} (no limit).
#' @param global_objects A character vector containing the name of R objects to be
#'   saved in a .RData file and loaded on each cluster node prior to calling
#'   \code{f}.
#' @param pkgs A character vector containing the names of packages that must
#'   be loaded on each cluster node. By default, it includes all packages
#'   loaded by the user when \code{slurm_map} is called.
#' @param libPaths A character vector describing the location of additional R
#'   library trees to search through, or NULL. The default value of NULL
#'   corresponds to libraries returned by \code{.libPaths()} on a cluster node.
#'   Non-existent library trees are silently ignored.
#' @param rscript_path The location of the Rscript command. If not specified, 
#'   defaults to the location of Rscript within the R installation being run.
#' @param r_template The path to the template file for the R script run on each node. 
#'   If NULL, uses the default template "rslurm/templates/slurm_run_R.txt".
#' @param sh_template The path to the template file for the sbatch submission script. 
#'   If NULL, uses the default template "rslurm/templates/submit_sh.txt".
#' @param slurm_options A named list of options recognized by \code{sbatch}; see
#'   Details below for more information.
#' @param submit Whether or not to submit the job to the cluster with
#'   \code{sbatch}; see Details below for more information.
#' @return A \code{slurm_job} object containing the \code{jobname} and the
#'   number of \code{nodes} effectively used.
#' @seealso \code{\link{slurm_call}} to evaluate a single function call.
#' @seealso \code{\link{slurm_apply}} to evaluate a function row-wise over a
#' data frame of parameters.
#' @seealso \code{\link{cancel_slurm}}, \code{\link{cleanup_files}},
#'   \code{\link{get_slurm_out}} and \code{\link{get_job_status}}
#'   which use the output of this function.
#' @examples
#' \dontrun{
#' sjob <- slurm_map(func, list)
#' get_job_status(sjob) # Prints console/error output once job is completed.
#' func_result <- get_slurm_out(sjob, "table") # Loads output data into R.
#' cleanup_files(sjob)
#' }
#' @export
slurm_map <- function(x, f, ..., jobname = NA, nodes = 2,
                      cpus_per_node = 2, processes_per_node = cpus_per_node,
                      preschedule_cores = TRUE, job_array_task_limit = NULL, global_objects = NULL, 
                      pkgs = rev(.packages()), libPaths = NULL, 
                      rscript_path = NULL, r_template = NULL, sh_template = NULL, 
                      slurm_options = list(), submit = TRUE) {
    # Check inputs
    if (!is.list(x)) {
        stop("first argument to slurm_map should be a list")
    }
    if (!is.function(f)) {
        stop("second argument to slurm_map should be a function")
    }
    if (!is.numeric(nodes) || length(nodes) != 1) {
        stop("nodes should be a single number")
    }
    if (!is.numeric(cpus_per_node) || length(cpus_per_node) != 1) {
        stop("cpus_per_node should be a single number")
    }
    
    # Default templates
    if(is.null(r_template)) {
        r_template <- system.file("templates/slurm_map_R.txt", package = "rslurm")
    }
    if(is.null(sh_template)) {
        sh_template <- system.file("templates/submit_sh.txt", package = "rslurm")
    }

    jobname <- make_jobname(jobname)

    # Create temp folder
    tmpdir <- paste0("_rslurm_", jobname)
    dir.create(tmpdir, showWarnings = FALSE)
    
    # Unpack additional arguments
    more_args <- list(...)

    saveRDS(x, file = file.path(tmpdir, "x.RDS"))
    saveRDS(f, file = file.path(tmpdir, "f.RDS"))
    saveRDS(more_args, file = file.path(tmpdir, "more_args.RDS"))
    if (!is.null(global_objects)) {
        save(list = global_objects,
             file = file.path(tmpdir, "add_objects.RData"),
             envir = environment(f))
    }    
    
    # Get chunk size (nb. of list elements per node)
    # Special case if less list elements than CPUs in cluster
    if (length(x) < cpus_per_node * nodes) {
        nchunk <- cpus_per_node
    } else {
        nchunk <- ceiling(length(x) / nodes)
    }
    # Re-adjust number of nodes (only matters for small sets)
    nodes <- ceiling(length(x) / nchunk)

    # Create a R script to run function in parallel on each node
    template_r <- readLines(r_template)
    script_r <- whisker::whisker.render(template_r,
                    list(pkgs = pkgs,
                         add_obj = !is.null(global_objects),
                         nchunk = nchunk,
                         cpus_per_node = cpus_per_node,
                         processes_per_node = processes_per_node,
                         preschedule_cores = preschedule_cores,
                         libPaths = libPaths))
    writeLines(script_r, file.path(tmpdir, "slurm_run.R"))

    # Create submission bash script
    template_sh <- readLines(sh_template)
    slurm_options <- format_option_list(slurm_options)
    if (is.null(rscript_path)){
        rscript_path <- file.path(R.home("bin"), "Rscript")
    }
    script_sh <- whisker::whisker.render(template_sh,
                    list(max_node = nodes - 1,
                         job_array_task_limit = ifelse(is.null(job_array_task_limit), "", paste0("%", job_array_task_limit)),
                         cpus_per_node = cpus_per_node,
                         jobname = jobname,
                         flags = slurm_options$flags,
                         options = slurm_options$options,
                         rscript = rscript_path))
    writeLines(script_sh, file.path(tmpdir, "submit.sh"))

    # Submit job to Slurm if applicable
    if (submit && system('squeue', ignore.stdout = TRUE)) {
        submit <- FALSE
        cat("Cannot submit; no Slurm workload manager found\n")
    }
    if (submit) {
        jobid <- submit_slurm_job(tmpdir)
    } else {
        jobid <- NA
        cat(paste("Submission scripts output in directory", tmpdir,"\n"))
    }

    # Return 'slurm_job' object
    slurm_job(jobname, jobid, nodes)
}

library(rslurm)
context("slurm_map")

# Locates sinfo on system, returns "1" if not found.
SLURM = Sys.which("sinfo") == ""
SLURM_MSG = 'Only test on Slurm head node.'
SLURM_OPTS = list(time = '1')

Sys.setenv(R_TESTS = "")
    
set.seed(123)

# Create a list of paired mean/sd values for normal distributions
pars <- data.frame(par_m = 1:10,
                   par_sd = seq(0.1, 1, length.out = 10))
pars_list <- split(pars, 1:10)

# Create a function to parallelize
ftest <- function(pars, par_n = 10^6, ...) {
    samp <- rnorm(par_n, pars$par_m, pars$par_sd)
    c(s_m = mean(samp), s_sd = sd(samp))
}

# ## FIXME
# saveRDS(Sys.getenv(), 'testthat_env.RDS')
# slurm_apply(function (i) Sys.getenv(), data.frame(i = c(0)), pkgs = c(), jobname = 'test0', nodes = 1, cpus_per_node = 1)

test_that("slurm_map gives correct output", {
    skip_on_os("windows")
    if (SLURM) skip(SLURM_MSG)
    sjob <- slurm_map(pars_list, ftest, jobname = "test1", nodes = 2, 
                      cpus_per_node = 1, slurm_options = SLURM_OPTS)
    res <- get_slurm_out(sjob, "table")
    res_raw <- get_slurm_out(sjob, "raw")
    cleanup_files(sjob)
    expect_equal(pars, res, tolerance = 0.01, check.attributes = FALSE)
    expect_equal(pars, as.data.frame(do.call(rbind, res_raw)),
                 tolerance = 0.01, check.attributes = FALSE)
})

test_that("slurm_map works with single list element", {
    skip_on_os("windows")
    if (SLURM) skip(SLURM_MSG)
    sjob <- slurm_map(pars_list[1], ftest, nodes = 2, jobname = "test2",
                      cpus_per_node = 1, slurm_options = SLURM_OPTS)
    res <- get_slurm_out(sjob, "table")
    cleanup_files(sjob)
    expect_equal(sjob$nodes, 1)
    expect_equal(pars[1, ], res, tolerance = 0.01, check.attributes = FALSE)  
})

test_that("slurm_map correctly handles global_objects", {
    skip_on_os("windows")
    if (SLURM) skip(SLURM_MSG)
    sjob <- slurm_map(as.list(1:length(pars_list)),
                      function(i) ftest(pars_list[[i]]),
                      global_objects = c('ftest', 'pars_list'), jobname = "test3",
                      nodes = 2, cpus_per_node = 1,
                      slurm_options = SLURM_OPTS)
    res <- get_slurm_out(sjob, "table")
    cleanup_files(sjob)
    expect_equal(pars, res, tolerance = 0.01, check.attributes = FALSE)
})

test_that("slurm_map correctly handles arguments passed with ...", {
    skip_on_os("windows")
    if (SLURM) skip (SLURM_MSG)
    sjob <- slurm_map(pars_list, ftest,
                      par_n = 10^6, jobname = "test4",
                      nodes = 2, cpus_per_node = 1,
                      slurm_options = SLURM_OPTS)
    res <- get_slurm_out(sjob, "table")
    cleanup_files(sjob)
    expect_equal(pars, res, tolerance = 0.01, check.attributes = FALSE)
    
})

library(rslurm)
context("local slurm_map")

Sys.setenv(R_TESTS = "")
    
set.seed(123)

# Create a list of mean/sd values for normal distributions
pars <- data.frame(par_m = 1:10,
                   par_sd = seq(0.1, 1, length.out = 10))
pars_list <- split(pars, 1:10)

# Create a function to parallelize
ftest <- function(pars, par_n = 10^6, ...) {
    samp <- rnorm(par_n, pars$par_m, pars$par_sd)
    c(s_m = mean(samp), s_sd = sd(samp))
}


# Test slurm_apply locally

msg <- capture.output(
    sjob1 <- slurm_map(pars_list, ftest, jobname = "test1",
                       nodes = 2, cpus_per_node = 1, submit = FALSE)
)

test_that("slurm_map gives correct output", {
    skip_on_os("windows")
    sjob1 <- local_slurm_array(sjob1)
    res <- get_slurm_out(sjob1, "table", wait = FALSE)
    res_raw <- get_slurm_out(sjob1, "raw", wait = FALSE)
    expect_equal(pars, res, tolerance = 0.01, check.attributes = FALSE)
    expect_equal(pars, as.data.frame(do.call(rbind, res_raw)),
                 tolerance = 0.01, check.attributes = FALSE)
})


# Test for degenerate case (length 1 list)

msg <- capture.output(
    sjob2 <- slurm_map(pars_list[1], ftest, jobname = "test2", 
                       nodes = 2, cpus_per_node = 1, submit = FALSE)
)

test_that("slurm_map works with length 1 list", {
    skip_on_os("windows")
    sjob2 <- local_slurm_array(sjob2)
    res <- get_slurm_out(sjob2, "table", wait = FALSE)
    expect_equal(pars[1, ], res, tolerance = 0.01, check.attributes = FALSE)  
})

# Test slurm_map with global_objects

msg <- capture.output(
    sjob3 <- slurm_map(as.list(1:length(pars_list)),
                       function(i) ftest(pars_list[[i]]),
                       global_objects = c('ftest', 'pars_list'),
                       jobname = "test3",
                       nodes = 2, cpus_per_node = 1, submit = FALSE)
)

test_that("slurm_map correctly handles global_objects", {
    skip_on_os("windows")
    sjob3 <- local_slurm_array(sjob3)
    res <- get_slurm_out(sjob3, "table", wait = FALSE)
    expect_equal(pars, res, tolerance = 0.01, check.attributes = FALSE)
})

# Test slurm_map with dots as additional arguments

msg <- capture.output(
    sjob4 <- slurm_map(pars_list, ftest, par_n = 10^6, jobname = "test4", 
                       nodes = 2, cpus_per_node = 1, submit = FALSE)
)

test_that("slurm_map correctly handles arguments passed with ...", {
    skip_on_os("windows")
    sjob4 <- local_slurm_array(sjob4)
    res <- get_slurm_out(sjob4, "table", wait = FALSE)
    expect_equal(pars$par_m, res$s_m, tolerance = 0.01)  
})

# Cleanup all temporary files at the end
# Pause to make sure folders are free to be deleted
Sys.sleep(1)
lapply(list(sjob1, sjob2, sjob3, sjob4), cleanup_files, wait = FALSE)

#' Execute a Slurm job locally
#' 
#' Run a previously created \code{slurm_job} object locally instead of on a
#' Slurm cluster
#' 
#' This function is most useful for testing your function on a reduced dataset
#' before submitting the full job to the Slurm cluster.
#' 
#' Call \code{local_slurm_array} on a \code{slurm_job} object created with
#' \code{slurm_apply(..., submit = FALSE)} or \code{slurm_map(..., submit = FALSE)}.
#' The job will run serially on the local system rather than being submitted
#' to the Slurm cluster. 
#' 
#' @param slr_job An object of class \code{slurm_job}.
#' @param rscript_path The location of the Rscript command. If not specified, 
#'   defaults to the location of Rscript within the R installation being run.
#'  
#' @examples
#' \dontrun{
#' sjob <- slurm_apply(func, pars, submit = FALSE)
#' local_slurm_array(sjob)
#' func_result <- get_slurm_out(sjob, "table") # Loads output data into R.
#' cleanup_files(sjob)
#' }
#' @export
local_slurm_array <- function(slr_job, rscript_path = NULL) {
    # Check that input is a slurm_job
    if (!inherits(slr_job, 'slurm_job')) {
        stop("argument of local_slurm_array should be a slurm_job")
    }
    
    if (is.null(rscript_path)) {
        rscript_path <- file.path(R.home("bin"), "Rscript")
    }
    
    slurm_run_path <- paste0("_rslurm_", slr_job$jobname)
    
    # Execute each array task locally and sequentially
    for (i in 1:slr_job$nodes - 1) {
        system(paste0('cd ', slurm_run_path, ' && SLURM_ARRAY_TASK_ID=', i, ' ', rscript_path, ' --vanilla slurm_run.R'))
    }
    
    return(slr_job)
}

library(rslurm)
context("local slurm_apply")

Sys.setenv(R_TESTS = "")
    
set.seed(123)

# Create a data frame of mean/sd values for normal distributions
pars <- data.frame(par_m = 1:10,
                   par_sd = seq(0.1, 1, length.out = 10))

# Create a function to parallelize
ftest <- function(par_m, par_sd = 1, ...) {
    samp <- rnorm(10^6, par_m, par_sd)
    c(s_m = mean(samp), s_sd = sd(samp))
}

# Alternative version with n as argument, to test more_args
ftest2 <- function(par_m, par_sd = 1, par_n = 10^6, ...) {
    samp <- rnorm(par_n, par_m, par_sd)
    c(s_m = mean(samp), s_sd = sd(samp))
}

# Test slurm_apply locally

msg <- capture.output(
    sjob1 <- slurm_apply(ftest, pars, jobname = "test1", nodes = 2, 
                         cpus_per_node = 1, submit = FALSE)
)

test_that("slurm_apply gives correct output", {
    skip_on_os("windows")
    sjob1 <- local_slurm_array(sjob1)
    res <- get_slurm_out(sjob1, "table", wait = FALSE)
    res_raw <- get_slurm_out(sjob1, "raw", wait = FALSE)
    expect_equal(pars, res, tolerance = 0.01, check.attributes = FALSE)
    expect_equal(pars, as.data.frame(do.call(rbind, res_raw)),
                 tolerance = 0.01, check.attributes = FALSE)
})


# Test for degenerate cases (single parameter and/or single row)

msg <- capture.output(
    sjob2 <- slurm_apply(ftest, pars[, 1, drop = FALSE], jobname = "test2", 
                         nodes = 2, cpus_per_node = 1, submit = FALSE)
)

test_that("slurm_apply works with single parameter", {
    skip_on_os("windows")
    sjob2 <- local_slurm_array(sjob2)
    res <- get_slurm_out(sjob2, "table", wait = FALSE)
    expect_equal(pars$par_m, res$s_m, tolerance = 0.01)  
})

msg <- capture.output(
    sjob3 <- slurm_apply(ftest, pars[1, ], nodes = 2, jobname = "test3",
                         cpus_per_node = 1, submit = FALSE)
)

test_that("slurm_apply works with single row", {
    skip_on_os("windows")
    sjob3 <- local_slurm_array(sjob3)
    res <- get_slurm_out(sjob3, "table", wait = FALSE)
    expect_equal(sjob3$nodes, 1)
    expect_equal(pars[1, ], res, tolerance = 0.01, check.attributes = FALSE)  
})

msg <- capture.output(
    sjob4 <- slurm_apply(ftest, pars[1, 1, drop = FALSE], jobname = "test4",
                         nodes = 2, cpus_per_node = 1, submit = FALSE)
)

test_that("slurm_apply works with single parameter and single row", {
    skip_on_os("windows")
    sjob4 <- local_slurm_array(sjob4)
    res <- get_slurm_out(sjob4, "table", wait = FALSE)
    expect_equal(pars$par_m[1], res$s_m, tolerance = 0.01)  
})

# Test slurm_apply with global_objects

msg <- capture.output(
    sjob5 <- slurm_apply(function(i) ftest(pars[i, 1], pars[i, 2]),
                         data.frame(i = 1:nrow(pars)),
                         global_objects = c('ftest', 'pars'), jobname = "test5",
                         nodes = 2, cpus_per_node = 1, submit = FALSE)
)

test_that("slurm_apply correctly handles global_objects", {
    skip_on_os("windows")
    sjob5 <- local_slurm_array(sjob5)
    res <- get_slurm_out(sjob5, "table", wait = FALSE)
    expect_equal(pars, res, tolerance = 0.01, check.attributes = FALSE)
})

# Test slurm_apply with ... arguments

msg <- capture.output(
    sjob6 <- slurm_apply(ftest2,
                         pars,
                         par_n = 10^6,
                         global_objects = c('ftest2', 'pars'), jobname = "test6",
                         nodes = 2, cpus_per_node = 1, submit = FALSE)
)

test_that("slurm_apply correctly handles arguments passed with ...", {
    skip_on_os("windows")
    sjob6 <- local_slurm_array(sjob6)
    res <- get_slurm_out(sjob6, "table", wait = FALSE)
    expect_equal(pars, res, tolerance = 0.01, check.attributes = FALSE)
})

# Cleanup all temporary files at the end
# Pause to make sure folders are free to be deleted
Sys.sleep(1)
lapply(list(sjob1, sjob2, sjob3, sjob4, sjob5, sjob6), cleanup_files, wait = FALSE)

library(rslurm)
context("local slurm_call")

Sys.setenv(R_TESTS = "")

# Test slurm_call locally

msg <- capture.output({
    z <- 0
    sjob <- slurm_call(function(x, y) x * 2 + y + z, list(x = 5, y = 6),
                       global_objects = c('z'),
                       jobname = "test^\\* call", submit = FALSE)
})

test_that("slurm_job name is correctly edited", {
    expect_equal(sjob$jobname, "test_call")
})


test_that("slurm_call returns correct output", {
    skip_on_os("windows")
    sjob <- local_slurm_array(sjob)
    res <- get_slurm_out(sjob, wait = FALSE)
    expect_equal(res, 16)
})

# Pause to make sure temporary folder is free to be deleted
Sys.sleep(1)
cleanup_files(sjob, wait = FALSE)

#' Create a slurm_job object
#' 
#' This function creates a \code{slurm_job} object which can be passed to other
#' functions such as \code{\link{cancel_slurm}}, \code{\link{cleanup_files}}, 
#' \code{\link{get_slurm_out}} and \code{\link{get_job_status}}. 
#' 
#' In general, \code{slurm_job} objects are created automatically as the output of 
#' \code{\link{slurm_apply}} or \code{\link{slurm_call}}, but it may be necessary 
#' to manually recreate one if the job was submitted in a different R session.
#' 
#' @param jobname The name of the Slurm job. The rslurm-generated scripts and 
#' output files associated with a job should be found in the 
#' \emph{_rslurm_[jobname]} folder.
#' @param jobid The id of the Slurm job created by the sbatch command.
#' @param nodes The number of cluster nodes used by that job.
#' @return A \code{slurm_job} object.
#' @export
slurm_job <- function(jobname = NULL, jobid = NULL, nodes = NULL) {
    slr_job <- list(jobname = jobname, jobid = jobid, nodes = nodes)
    class(slr_job) <- "slurm_job"
    slr_job
}

#' Reads the output of a function calculated on the Slurm cluster 
#'
#' This function reads all function output files (one by cluster node used) from 
#' the specified Slurm job and returns the result in a single data frame
#' (if "table" format selected) or list (if "raw" format selected). It doesn't 
#' record any messages (including warnings or errors) output to the R console
#' during the computation; these can be consulted by invoking 
#' \code{\link{print_job_status}}.
#' 
#' The \code{outtype} option is only relevant for jobs submitted with 
#' \code{slurm_apply}. Jobs sent with \code{slurm_call} only return a single
#' object, and setting \code{outtype = "table"} creates an error in that case.
#' 
#' @param slr_job A \code{slurm_job} object.
#' @param outtype Can be "table" or "raw", see "Value" below for details.
#' @param wait Specify whether to block until \code{slr_job} completes.
#' @param ncores (optional) If not null, the number of cores passed to
#' mclapply
#' @return If \code{outtype = "table"}: A data frame with one column by 
#'   return value of the function passed to \code{slurm_apply}, where 
#'   each row is the output of the corresponding row in the params data frame 
#'   passed to \code{slurm_apply}.
#'   
#'   If \code{outtype = "raw"}: A list where each element is the output 
#'   of the function passed to \code{slurm_apply} for the corresponding
#'   row in the params data frame passed to \code{slurm_apply}.
#' @seealso \code{\link{slurm_apply}}, \code{\link{slurm_call}}
#' @importFrom parallel mclapply
#' @export
get_slurm_out <- function(slr_job, outtype = "raw", wait = TRUE, 
                          ncores = NULL) {
    
    # Check arguments
    if (!(inherits(slr_job, "slurm_job"))) {
        stop("slr_job must be a slurm_job")
    }
    outtypes <- c("table", "raw")
    if (!(outtype %in% outtypes)) {
        stop(paste("outtype should be one of:", paste(outtypes, collapse = ', ')))
    }
    if (!(is.null(ncores) || (is.numeric(ncores) && length(ncores) == 1))) {
        stop("ncores must be an integer number of cores")
    }

    # Wait for slr_job using Slurm dependency
    if (wait) {
        wait_for_job(slr_job)
    }
    
    res_files <- paste0("results_", 0:(slr_job$nodes - 1), ".RDS")
    tmpdir <- paste0("_rslurm_", slr_job$jobname)
    missing_files <- setdiff(res_files, dir(path = tmpdir))
    if (length(missing_files) > 0) {
        missing_list <- paste(missing_files, collapse = ", ")
        warning(paste("The following files are missing:", missing_list))
    }
    res_files <- file.path(tmpdir, setdiff(res_files, missing_files))
    if (length(res_files) == 0) return(NA)
    
    if (is.null(ncores)) {
        slurm_out <- lapply(res_files, readRDS)
    } else {
        slurm_out <- mclapply(res_files, readRDS, mc.cores = ncores)
    }
    slurm_out <- do.call(c, slurm_out)
    
    if (outtype == "table") {
        slurm_out <- as.data.frame(do.call(rbind, slurm_out))
    }
    slurm_out
}

library(rslurm)
context("slurm_call")

# Locates sinfo on system, returns "1" if not found.
SLURM = Sys.which("sinfo") == ""
SLURM_MSG = 'Only test on Slurm head node.'
SLURM_OPTS = list(time = '1')

Sys.setenv(R_TESTS = "")

test_that("slurm_job name is correctly edited and output is correct", {
    skip_on_os("windows")
    if (SLURM) skip(SLURM_MSG)
    z <- 0
    sjob <- slurm_call(function(x, y) x * 2 + y + z, list(x = 5, y = 6),
                       global_objects = c('z'),
                       jobname = "test^\\* call", slurm_options = SLURM_OPTS)
    res <- get_slurm_out(sjob)
    cleanup_files(sjob)
    expect_equal(sjob$jobname, "test_call")
    expect_equal(res, 16)
})

test_that("slurm_call will handle a bytecoded function", {
    # generated in response to issue #14
    skip_on_os("windows")
    if (SLURM) skip(SLURM_MSG)
    params <- list(
        data = data.frame(
            x = seq(0, 1, by = 0.01),
            y = 0:100),
        formula = y ~ x)
    result_local <- do.call(lm, params)
    sjob <- slurm_call(lm, params, slurm_options = SLURM_OPTS)
    result_slurm <- get_slurm_out(sjob)
    cleanup_files(sjob)
    expect_equal(result_slurm$coefficients, result_local$coefficients)
})

test_that("slurm_call executes if no parameters provided", {
    skip_on_os("windows")
    if (SLURM) skip(SLURM_MSG)
    ftest <- function() sum((1:1000)^2)
    result_local <- ftest()
    sjob <- slurm_call(ftest, slurm_options = SLURM_OPTS)
    result_slurm <- get_slurm_out(sjob)
    cleanup_files(sjob)
    expect_equal(result_slurm, result_local)
})

library(rslurm)
context("slurm_apply")

# Locates sinfo on system, returns "1" if not found.
SLURM = Sys.which("sinfo") == ""
SLURM_MSG = 'Only test on Slurm head node.'
SLURM_OPTS = list(time = '1')

Sys.setenv(R_TESTS = "")
    
set.seed(123)

# Create a data frame of mean/sd values for normal distributions
pars <- data.frame(par_m = 1:10,
                   par_sd = seq(0.1, 1, length.out = 10))

# Create a function to parallelize
ftest <- function(par_m, par_sd = 1, ...) {
    samp <- rnorm(10^6, par_m, par_sd)
    c(s_m = mean(samp), s_sd = sd(samp))
}

# Alternative version with n as argument, to test more_args
ftest2 <- function(par_m, par_sd = 1, par_n = 10^6, ...) {
    samp <- rnorm(par_n, par_m, par_sd)
    c(s_m = mean(samp), s_sd = sd(samp))
}

# ## FIXME
# saveRDS(Sys.getenv(), 'testthat_env.RDS')
# slurm_apply(function (i) Sys.getenv(), data.frame(i = c(0)), pkgs = c(), jobname = 'test0', nodes = 1, cpus_per_node = 1)

test_that("slurm_apply gives correct output", {
    skip_on_os("windows")
    if (SLURM) skip(SLURM_MSG)
    sjob <- slurm_apply(ftest, pars, jobname = "test1", nodes = 2, 
                        cpus_per_node = 1, slurm_options = SLURM_OPTS)
    res <- get_slurm_out(sjob, "table")
    res_raw <- get_slurm_out(sjob, "raw")
    cleanup_files(sjob)
    expect_equal(pars, res, tolerance = 0.01, check.attributes = FALSE)
    expect_equal(pars, as.data.frame(do.call(rbind, res_raw)),
                 tolerance = 0.01, check.attributes = FALSE)
})

test_that("slurm_apply works with single parameter", {
    skip_on_os("windows")
    if (SLURM) skip(SLURM_MSG)
    sjob <- slurm_apply(ftest, pars[, 1, drop = FALSE], jobname = "test2", 
                        nodes = 2, cpus_per_node = 1, slurm_options = SLURM_OPTS)
    res <- get_slurm_out(sjob, "table")
    cleanup_files(sjob)
    expect_equal(pars$par_m, res$s_m, tolerance = 0.01)  
})

test_that("slurm_apply works with single row", {
    skip_on_os("windows")
    if (SLURM) skip(SLURM_MSG)
    sjob <- slurm_apply(ftest, pars[1, ], nodes = 2, jobname = "test3",
                        cpus_per_node = 1, slurm_options = SLURM_OPTS)
    res <- get_slurm_out(sjob, "table")
    cleanup_files(sjob)
    expect_equal(sjob$nodes, 1)
    expect_equal(pars[1, ], res, tolerance = 0.01, check.attributes = FALSE)  
})

test_that("slurm_apply works with single parameter and single row", {
    skip_on_os("windows")
    if (SLURM) skip(SLURM_MSG)
    sjob <- slurm_apply(ftest, pars[1, 1, drop = FALSE], jobname = "test4",
                        nodes = 2, cpus_per_node = 1,
                        slurm_options = SLURM_OPTS)
    res <- get_slurm_out(sjob, "table")
    cleanup_files(sjob)
    expect_equal(pars$par_m[1], res$s_m, tolerance = 0.01)  
})

test_that("slurm_apply correctly handles global_objects", {
    skip_on_os("windows")
    if (SLURM) skip(SLURM_MSG)
    sjob <- slurm_apply(function(i) ftest(pars[i, 1], pars[i, 2]),
                        data.frame(i = 1:nrow(pars)),
                        global_objects = c('ftest', 'pars'), jobname = "test5",
                        nodes = 2, cpus_per_node = 1,
                        slurm_options = SLURM_OPTS)
    res <- get_slurm_out(sjob, "table")
    cleanup_files(sjob)
    expect_equal(pars, res, tolerance = 0.01, check.attributes = FALSE)
})

test_that("slurm_apply correctly handles arguments given as dots", {
    skip_on_os("windows")
    if (SLURM) skip (SLURM_MSG)
    sjob <- slurm_apply(ftest2,
                        pars,
                        par_n = 10^6,
                        global_objects = c('ftest2', 'pars'), jobname = "test6",
                        nodes = 2, cpus_per_node = 1,
                        slurm_options = SLURM_OPTS)
    res <- get_slurm_out(sjob, "table")
    cleanup_files(sjob)
    expect_equal(pars, res, tolerance = 0.01, check.attributes = FALSE)
    
})

test_that("job_array_task_limit argument works correctly", {
    skip_on_os("windows")
    if (SLURM) skip(SLURM_MSG)
    sjob <- slurm_apply(ftest, pars, jobname = "test1", nodes = 2, 
                        cpus_per_node = 1, slurm_options = SLURM_OPTS,
                        job_array_task_limit = 1)
    res <- get_slurm_out(sjob, "table")
    res_raw <- get_slurm_out(sjob, "raw")
    cleanup_files(sjob)
    expect_equal(pars, res, tolerance = 0.01, check.attributes = FALSE)
    expect_equal(pars, as.data.frame(do.call(rbind, res_raw)),
                 tolerance = 0.01, check.attributes = FALSE)
})

#' Introduction to the \code{rslurm} Package
#' 
#' Send long-running or parallel jobs to a Slurm workload manager (i.e. cluster)
#' using the \code{\link{slurm_call}}, \code{\link{slurm_apply}}, or
#' \code{\link{slurm_map}} functions.
#' 
#' @section Job submission:
#'   
#'   This package includes three core functions used to send computations to a 
#'   Slurm cluster: 1) \code{\link{slurm_call}} executes a function using a 
#'   single set of parameters (passed as a list), 2) \code{\link{slurm_apply}} 
#'   evaluates a function in parallel for each row of parameters in a given 
#'   data frame, and 3) \code{\link{slurm_map}} evaluates a function in parallel
#'   for each element of a list. 
#'   The functions \code{slurm_apply} and \code{slurm_map} automatically split the 
#'   parameter rows or list elements into equal-size chunks, 
#'   each chunk to be processed by a separate cluster node. 
#'   They use functions from the \code{\link[parallel]{parallel-package}} 
#'   package to parallelize computations across processors on a given node.
#'   
#'   The output of \code{slurm_apply}, \code{slurm_map}, or \code{slurm_call} 
#'   is a \code{slurm_job} object that serves as an input to the other functions in the package: 
#'   \code{\link{print_job_status}}, \code{\link{cancel_slurm}}, 
#'   \code{\link{get_slurm_out}} and \code{\link{cleanup_files}}.
#'   
#' @section Function specification:
#'   
#'   To be compatible with \code{\link{slurm_apply}}, a function may accept any 
#'   number of single value parameters. The names of these parameters must match
#'   the column names of the \code{params} data frame supplied. There are no 
#'   restrictions on the types of parameters passed as a list to 
#'   \code{\link{slurm_call}} or \code{\link{slurm_map}}
#'   
#'   If the function passed to \code{slurm_call} or \code{slurm_apply} requires 
#'   knowledge of any R objects (data, custom helper functions) besides 
#'   \code{params}, a character vector corresponding to their names should be 
#'   passed to the optional \code{global_objects} argument.
#'   
#'   When parallelizing a function, since any error will interrupt all 
#'   calculations for the current node, it may be useful to wrap expressions 
#'   which may generate errors into a \code{\link[base]{try}} or 
#'   \code{\link[base:conditions]{tryCatch}} function. This will ensure the computation 
#'   continues with the next parameter set after reporting the error.
#'   
#' @section Output Format:
#'   
#'   The default output format for \code{get_slurm_out} (\code{outtype = "raw"})
#'   is a list where each element is the return value of one function call. If 
#'   the function passed to \code{slurm_apply} produces a vector output, you may
#'   use \code{outtype = "table"} to collect the output in a single data frame, 
#'   with one row by function call.
#'
#' @section Slurm Configuration:
#'   
#'   Advanced options for the Slurm workload manager may accompany job submission
#'   by \code{\link{slurm_call}}, \code{\link{slurm_map}}, and \code{\link{slurm_apply}} 
#'   through the optional \code{slurm_options} argument. For example, passing
#'   \code{list(time = '1:30:00')} for this options limits the job to 1 hour and 30
#'   minutes. Some advanced configuration must be set through environment 
#'   variables. On a multi-cluster head node, for example, the \code{SLURM_CLUSTERS}
#'   environment variable must be set to direct jobs to a non-default cluster.
#'   
#' @examples
#' 
#' \dontrun{
#' # Create a data frame of mean/sd values for normal distributions 
#' pars <- data.frame(par_m = seq(-10, 10, length.out = 1000), 
#'                    par_sd = seq(0.1, 10, length.out = 1000))
#'                    
#' # Create a function to parallelize
#' ftest <- function(par_m, par_sd) {
#'  samp <- rnorm(10^7, par_m, par_sd)
#'  c(s_m = mean(samp), s_sd = sd(samp))
#' }
#' 
#' sjob1 <- slurm_apply(ftest, pars)
#' print_job_status(sjob1)
#' res <- get_slurm_out(sjob1, "table")
#' all.equal(pars, res) # Confirm correct output
#' cleanup_files(sjob1)
#' }
#'   
#' @importFrom utils capture.output
#' @docType package
#' @name rslurm-package
#'   
NULL

#' Cancels a scheduled Slurm job
#'
#' This function cancels the specified Slurm job by invoking the Slurm 
#' \code{scancel} command. It does \emph{not} delete the temporary files 
#' (e.g. scripts) created by \code{\link{slurm_apply}} or 
#' \code{\link{slurm_call}}. Use \code{\link{cleanup_files}} to remove those 
#' files. 
#' 
#' @param slr_job A \code{slurm_job} object.
#' @seealso \code{\link{cleanup_files}}
#' @export
cancel_slurm <- function(slr_job) {
    if (!(inherits(slr_job, "slurm_job"))) stop("input must be a slurm_job")
    system(paste("scancel -n", slr_job$jobname))
}


#' Deletes temporary files associated with a Slurm job 
#'
#' This function deletes all temporary files associated with the specified Slurm
#' job, including files created by \code{\link{slurm_apply}} or 
#' \code{\link{slurm_call}}, as well as outputs from the cluster. These files
#' should be located in the \emph{_rslurm_[jobname]} folder of the current
#' working directory.
#' 
#' @param slr_job A \code{slurm_job} object.
#' @param wait Specify whether to block until \code{slr_job} completes.
#' @examples 
#' \dontrun{
#' sjob <- slurm_apply(func, pars)
#' print_job_status(sjob) # Prints console/error output once job is completed.
#' func_result <- get_slurm_out(sjob, "table") # Loads output data into R.
#' cleanup_files(sjob)
#' }
#' @seealso \code{\link{slurm_apply}}, \code{\link{slurm_call}}
#' @export
cleanup_files <- function(slr_job, wait = TRUE) {
    if (!(inherits(slr_job, "slurm_job"))) stop("input must be a slurm_job")
    if (wait) wait_for_job(slr_job)
    tmpdir <- paste0("_rslurm_", slr_job$jobname)
    if (!(tmpdir %in% dir())) stop(paste("folder", tmpdir, "not found"))
    unlink(tmpdir, recursive = TRUE)
}

#' Parallel execution of a function on the Slurm cluster
#'
#' Use \code{slurm_apply} to compute function over multiple sets of
#' parameters in parallel, spread across multiple nodes of a Slurm cluster.
#'
#' This function creates a temporary folder ("_rslurm_[jobname]") in the current
#' directory, holding .RData and .RDS data files, the R script to run and the Bash
#' submission script generated for the Slurm job.
#'
#' The set of input parameters is divided in equal chunks sent to each node, and
#' \code{f} is evaluated in parallel within each node using functions from the
#' \code{parallel} R package. The names of any other R objects (besides
#' \code{params}) that \code{f} needs to access should be included in
#' \code{add_objects}.
#'
#' Use \code{slurm_options} to set any option recognized by \code{sbatch}, e.g.
#' \code{slurm_options = list(time = "1:00:00", share = TRUE)}.
#' See \url{http://slurm.schedmd.com/sbatch.html} for details on possible options.
#' Note that full names must be used (e.g. "time" rather than "t") and that flags
#' (such as "share") must be specified as TRUE. The "array", "job-name", "nodes", 
#' "cpus-per-task" and "output" options are already determined by 
#' \code{slurm_apply} and should not be manually set.
#'
#' When processing the computation job, the Slurm cluster will output two types
#' of files in the temporary folder: those containing the return values of the
#' function for each subset of parameters ("results_[node_id].RDS") and those
#' containing any console or error output produced by R on each node
#' ("slurm_[node_id].out").
#'
#' If \code{submit = TRUE}, the job is sent to the cluster and a confirmation
#' message (or error) is output to the console. If \code{submit = FALSE},
#' a message indicates the location of the saved data and script files; the
#' job can be submitted manually by running the shell command
#' \code{sbatch submit.sh} from that directory.
#'
#' After sending the job to the Slurm cluster, \code{slurm_apply} returns a
#' \code{slurm_job} object which can be used to cancel the job, get the job
#' status or output, and delete the temporary files associated with it. See
#' the description of the related functions for more details.
#'
#' @param f A function that accepts one or many single values as parameters and
#'   may return any type of R object.
#' @param params A data frame of parameter values to apply \code{f} to. Each
#'   column corresponds to a parameter of \code{f} (\emph{Note}: names must
#'   match) and each row corresponds to a separate function call.
#' @param jobname The name of the Slurm job; if \code{NA}, it is assigned a
#'   random name of the form "slr####".
#' @param nodes The (maximum) number of cluster nodes to spread the calculation
#'   over. \code{slurm_apply} automatically divides \code{params} in chunks of
#'   approximately equal size to send to each node. Less nodes are allocated if
#'   the parameter set is too small to use all CPUs on the requested nodes.
#' @param cpus_per_node The number of CPUs requested per node, i.e., how many
#'   processes to run in parallel per node. This argument is mapped to the
#'   Slurm parameter \code{cpus-per-task}.
#' @param preschedule_cores Corresponds to the \code{mc.preschedule} argument of 
#' \code{parallel::mcmapply}. Defaults to \code{TRUE}. If \code{TRUE}, the 
#' jobs are assigned to cores before computation. If \code{FALSE}, a new job is
#' created for each row of \code{params}. Setting \code{FALSE} may be faster if 
#' different values of \code{params} result in very variable completion time for
#' jobs.
#' @param add_objects A character vector containing the name of R objects to be
#'   saved in a .RData file and loaded on each cluster node prior to calling
#'   \code{f}.
#' @param pkgs A character vector containing the names of packages that must
#'   be loaded on each cluster node. By default, it includes all packages
#'   loaded by the user when \code{slurm_apply} is called.
#' @param libPaths A character vector describing the location of additional R
#'   library trees to search through, or NULL. The default value of NULL
#'   corresponds to libraries returned by \code{.libPaths()} on a cluster node.
#'   Non-existent library trees are silently ignored.
#' @param rscript_path The location of the Rscript command. If not specified, 
#'   defaults to the location of Rscript within the R installation being run.
#' @param r_template The path to the template file for the R script run on each node. 
#'   If NULL, uses the default template "rslurm/templates/slurm_run_R.txt".
#' @param sh_template The path to the template file for the sbatch submission script. 
#'   If NULL, uses the default template "rslurm/templates/submit_sh.txt".
#' @param slurm_options A named list of options recognized by \code{sbatch}; see
#'   Details below for more information.
#' @param submit Whether or not to submit the job to the cluster with
#'   \code{sbatch}; see Details below for more information.
#' @return A \code{slurm_job} object containing the \code{jobname} and the
#'   number of \code{nodes} effectively used.
#' @seealso \code{\link{slurm_call}} to evaluate a single function call.
#' @seealso \code{\link{cancel_slurm}}, \code{\link{cleanup_files}},
#'   \code{\link{get_slurm_out}} and \code{\link{get_job_status}}
#'   which use the output of this function.
#' @examples
#' \dontrun{
#' sjob <- slurm_apply(func, pars)
#' get_job_status(sjob) # Prints console/error output once job is completed.
#' func_result <- get_slurm_out(sjob, "table") # Loads output data into R.
#' cleanup_files(sjob)
#' }
#' @export
slurm_apply <- function(f, params, jobname = NA, 
                        nodes = 2, cpus_per_node = 2, preschedule_cores = TRUE,
                        add_objects = NULL, pkgs = rev(.packages()), libPaths = NULL, 
                        rscript_path = NULL, r_template = NULL, sh_template = NULL, 
                        slurm_options = list(), submit = TRUE) {
    # Check inputs
    if (!is.function(f)) {
        stop("first argument to slurm_apply should be a function")
    }
    if (!is.data.frame(params)) {
        stop("second argument to slurm_apply should be a data.frame")
    }
    if (is.null(names(params)) || any(!names(params) %in% names(formals(f)))) {
        stop("column names of params must match arguments of f")
    }
    if (!is.numeric(nodes) || length(nodes) != 1) {
        stop("nodes should be a single number")
    }
    if (!is.numeric(cpus_per_node) || length(cpus_per_node) != 1) {
        stop("cpus_per_node should be a single number")
    }
    
    # Default templates
    if(is.null(r_template)) {
        r_template <- system.file("templates/slurm_run_R.txt", package = "rslurm")
    }
    if(is.null(sh_template)) {
        sh_template <- system.file("templates/submit_sh.txt", package = "rslurm")
    }

    jobname <- make_jobname(jobname)

    # Create temp folder
    tmpdir <- paste0("_rslurm_", jobname)
    dir.create(tmpdir, showWarnings = FALSE)

    saveRDS(params, file = file.path(tmpdir, "params.RDS"))
    saveRDS(f, file = file.path(tmpdir, "f.RDS"))
    if (!is.null(add_objects)) {
        save(list = add_objects,
             file = file.path(tmpdir, "add_objects.RData"),
             envir = environment(f))
    }    
    
    # Get chunk size (nb. of param. sets by node)
    # Special case if less param. sets than CPUs in cluster
    if (nrow(params) < cpus_per_node * nodes) {
        nchunk <- cpus_per_node
    } else {
        nchunk <- ceiling(nrow(params) / nodes)
    }
    # Re-adjust number of nodes (only matters for small sets)
    nodes <- ceiling(nrow(params) / nchunk)

    # Create a R script to run function in parallel on each node
    template_r <- readLines(r_template)
    script_r <- whisker::whisker.render(template_r,
                    list(pkgs = pkgs,
                         add_obj = !is.null(add_objects),
                         nchunk = nchunk,
                         cpus_per_node = cpus_per_node,
                         preschedule_cores = preschedule_cores,
                         libPaths = libPaths))
    writeLines(script_r, file.path(tmpdir, "slurm_run.R"))

    # Create submission bash script
    template_sh <- readLines(sh_template)
    slurm_options <- format_option_list(slurm_options)
    if (is.null(rscript_path)){
        rscript_path <- file.path(R.home("bin"), "Rscript")
    }
    script_sh <- whisker::whisker.render(template_sh,
                    list(max_node = nodes - 1,
                         cpus_per_node = cpus_per_node,
                         jobname = jobname,
                         flags = slurm_options$flags,
                         options = slurm_options$options,
                         rscript = rscript_path))
    writeLines(script_sh, file.path(tmpdir, "submit.sh"))

    # Submit job to Slurm if applicable
    if (submit && system('squeue', ignore.stdout = TRUE)) {
        submit <- FALSE
        cat("Cannot submit; no Slurm workload manager found\n")
    }
    if (submit) {
        submit_slurm_job(tmpdir)
    } else {
        cat(paste("Submission scripts output in directory", tmpdir))
    }

    # Return 'slurm_job' object
    slurm_job(jobname, nodes)
}

#' Execution of a single function call on the Slurm cluster
#' 
#' Use \code{slurm_call} to perform a single function evaluation a the Slurm 
#' cluster.
#' 
#' This function creates a temporary folder ("_rslurm_[jobname]") in the current
#' directory, holding .RData and .RDS data files, the R script to run and the
#' Bash submission script generated for the Slurm job.
#' 
#' The names of any other R objects (besides \code{params}) that \code{f} needs 
#' to access should be listed in the \code{global_objects} argument.
#' 
#' Use \code{slurm_options} to set any option recognized by \code{sbatch}, e.g. 
#' \code{slurm_options = list(time = "1:00:00", share = TRUE)}. See
#' \url{http://slurm.schedmd.com/sbatch.html} for details on possible options. 
#' Note that full names must be used (e.g. "time" rather than "t") and that
#' flags (such as "share") must be specified as TRUE. The "job-name", "ntasks" 
#' and "output" options are already determined by \code{slurm_call} and should 
#' not be manually set.
#' 
#' When processing the computation job, the Slurm cluster will output two files 
#' in the temporary folder: one with the return value of the function 
#' ("results_0.RDS") and one containing any console or error output produced by
#' R ("slurm_[node_id].out").
#' 
#' If \code{submit = TRUE}, the job is sent to the cluster and a confirmation 
#' message (or error) is output to the console. If \code{submit = FALSE}, a
#' message indicates the location of the saved data and script files; the job
#' can be submitted manually by running the shell command \code{sbatch
#' submit.sh} from that directory.
#' 
#' After sending the job to the Slurm cluster, \code{slurm_call} returns a 
#' \code{slurm_job} object which can be used to cancel the job, get the job 
#' status or output, and delete the temporary files associated with it. See the
#' description of the related functions for more details.
#' 
#' @param f Any R function.
#' @param params A named list of parameters to pass to \code{f}.
#' @param jobname The name of the Slurm job; if \code{NA}, it is assigned a 
#'   random name of the form "slr####".
#' @param global_objects A character vector containing the name of R objects to be 
#'   saved in a .RData file and loaded on each cluster node prior to calling 
#'   \code{f}.
#' @param add_objects Older deprecated name of \code{global_objects}, retained for
#'   backwards compatibility.
#' @param pkgs A character vector containing the names of packages that must be
#'   loaded on each cluster node. By default, it includes all packages loaded by
#'   the user when \code{slurm_call} is called.
#' @param libPaths A character vector describing the location of additional R
#'   library trees to search through, or NULL. The default value of NULL
#'   corresponds to libraries returned by \code{.libPaths()} on a cluster node.
#'   Non-existent library trees are silently ignored.
#' @param rscript_path The location of the Rscript command. If not specified, 
#'   defaults to the location of Rscript within the R installation being run.
#' @param r_template The path to the template file for the R script run on each node. 
#'   If NULL, uses the default template "rslurm/templates/slurm_run_single_R.txt".
#' @param sh_template The path to the template file for the sbatch submission script. 
#'   If NULL, uses the default template "rslurm/templates/submit_single_sh.txt".
#' @param slurm_options A named list of options recognized by \code{sbatch}; see
#'   Details below for more information.
#' @param submit Whether or not to submit the job to the cluster with 
#'   \code{sbatch}; see Details below for more information.
#' @return A \code{slurm_job} object containing the \code{jobname} and the number
#'   of \code{nodes} effectively used.
#' @seealso \code{\link{slurm_apply}} to parallelize a function over a parameter
#'   set.
#' @seealso \code{\link{cancel_slurm}}, \code{\link{cleanup_files}}, 
#'   \code{\link{get_slurm_out}} and \code{\link{get_job_status}} which use
#'   the output of this function.
#' @export
slurm_call <- function(f, params = list(), jobname = NA, global_objects = NULL, add_objects = NULL, 
                       pkgs = rev(.packages()), libPaths = NULL, rscript_path = NULL,
                       r_template = NULL, sh_template = NULL, slurm_options = list(), 
                       submit = TRUE) {
    # Check inputs
    if (!is.function(f)) {
        stop("first argument to slurm_call should be a function")
    }
    if (!missing(params)) {
        if (!is.list(params)) {
            stop("second argument to slurm_call should be a list")
        }
        if (is.null(names(params)) || (!is.primitive(f) && !"..." %in% names(formals(f)) && any(!names(params) %in% names(formals(f))))) {
            stop("names of params must match arguments of f")
        }
    } 
    
    # Check for use of deprecated argument
    if (!missing("add_objects")) {
        warning("Argument add_objects is deprecated; use global_objects instead.", .call = FALSE)
        global_objects <- add_objects
    }
    
    # Default templates
    if(is.null(r_template)) {
        r_template <- system.file("templates/slurm_run_single_R.txt", package = "rslurm")
    }
    if(is.null(sh_template)) {
        sh_template <- system.file("templates/submit_single_sh.txt", package = "rslurm")
    }
        
    jobname <- make_jobname(jobname)
    
    # Create temp folder
    tmpdir <- paste0("_rslurm_", jobname)
    dir.create(tmpdir, showWarnings = FALSE)
    
    saveRDS(params, file = file.path(tmpdir, "params.RDS"))
    saveRDS(f, file = file.path(tmpdir, "f.RDS"))
    if (!is.null(global_objects)) {
        save(list = global_objects,
             file = file.path(tmpdir, "add_objects.RData"),
             envir = environment(f))
    }    
    
    # Create a R script to run function on cluster
    template_r <- readLines(r_template)
    script_r <- whisker::whisker.render(template_r,
                    list(pkgs = pkgs,
                         add_obj = !is.null(global_objects),
                         libPaths = libPaths))
    writeLines(script_r, file.path(tmpdir, "slurm_run.R"))
    
    # Create submission bash script
    template_sh <- readLines(sh_template)
    slurm_options <- format_option_list(slurm_options)
    if (is.null(rscript_path)){
        rscript_path <- file.path(R.home("bin"), "Rscript")
    }
    script_sh <- whisker::whisker.render(template_sh, 
                                         list(jobname = jobname,
                                              flags = slurm_options$flags, 
                                              options = slurm_options$options,
                                              rscript = rscript_path))
    writeLines(script_sh, file.path(tmpdir, "submit.sh"))
    
    # Submit job to Slurm if applicable
    if (submit && system('squeue', ignore.stdout = TRUE)) {
        submit <- FALSE
        cat("Cannot submit; no Slurm workload manager found\n")
    }
    if (submit) {
        jobid <- submit_slurm_job(tmpdir)
    } else {
        jobid <- NA
        cat(paste("Submission scripts output in directory", tmpdir,"\n"))
    }

    # Return 'slurm_job' object
    slurm_job(jobname, jobid, 1)
}

#' @title Deprecated functions in package \pkg{rslurm}.
#' @description The functions listed below are deprecated and will be defunct in
#'   the near future. When possible, alternative functions with similar
#'   functionality are also mentioned. Help pages for deprecated functions are
#'   available at \code{help("-deprecated")}.
#' @name rslurm-deprecated
#' @keywords internal
NULL

library(base, quietly = TRUE)
library(methods, quietly = TRUE)
library(datasets, quietly = TRUE)
library(utils, quietly = TRUE)
library(grDevices, quietly = TRUE)
library(graphics, quietly = TRUE)
library(stats, quietly = TRUE)
library(rslurm, quietly = TRUE)


.rslurm_func <- readRDS('f.RDS')
.rslurm_params <- readRDS('params.RDS')
.rslurm_more_args <- readRDS('more_args.RDS')
.rslurm_id <- as.numeric(Sys.getenv('SLURM_ARRAY_TASK_ID'))
.rslurm_istart <- .rslurm_id * 5 + 1
.rslurm_iend <- min((.rslurm_id + 1) * 5, nrow(.rslurm_params))
.rslurm_result <- do.call(parallel::mcmapply, c(
    FUN = .rslurm_func,
    .rslurm_params[.rslurm_istart:.rslurm_iend, , drop = FALSE],
    MoreArgs = list(.rslurm_more_args),
    mc.cores = 2,
    mc.preschedule = TRUE,
    SIMPLIFY = FALSE))

saveRDS(.rslurm_result, file = paste0('results_', .rslurm_id, '.RDS'))

#' Get the status of a Slurm job
#'
#' This function returns the completion status of a Slurm job, its queue status 
#' if any and log outputs. 
#' 
#' The \code{queue} element of the output is a data frame matching the output
#' of the Slurm \code{squeue} command for that job; it will only indicate portions
#' of job that are running or in queue. The \code{log} element is a
#' vector of the contents of console/error output files for each node where the 
#' job is running.
#' 
#' @importFrom utils read.table
#' @param slr_job A \code{slurm_job} object.
#' @return A list with three elements: \code{completed} is a logical value 
#'   indicating if all portions of the job have completed or stopped, \code{queue} 
#'   contains the information on job elements still in queue, and
#'   \code{log} contains the console/error logs. 
#' @export
get_job_status <- function(slr_job) {
    
    if (!(inherits(slr_job, "slurm_job"))) stop("input must be a slurm_job")
    
    # Get queue info
    squeue_out <- suppressWarnings(
        system(paste("squeue -n", slr_job$jobname), intern = TRUE)
    )
    queue <- read.table(text = squeue_out, header = TRUE)
    completed <- nrow(queue) == 0
    
    # Get output logs
    tmpdir <- paste0("_rslurm_", slr_job$jobname)
    out_files <- list.files(tmpdir, pattern = "slurm.*\\.out", full.names = TRUE)
    logs <- vapply(out_files, 
                   function(outf) paste(readLines(outf), collapse = "\n"),
                   "")
    
    job_status <- list(completed = completed, queue = queue, log = logs)
    class(job_status) <- "slurm_job_status"
    job_status
}  


# Format job status output
print.slurm_job_status <- function(stat) {
    if (stat$complete) {
        cat("Job completed or stopped.\n\n")
    } else {
        print(stat$queue)
        cat("\n")
    }
    
    cat("Last console output\n\n")
    shorten_log <- function(txt, lmax = 60) { # Shorten txt to lmax chars.
        l <- nchar(txt)
        ifelse(l <= lmax, txt,
               paste("...", substr(txt, l - lmax, l)))
    }
    
    for (i in seq_along(stat$log)) {
        cat(paste0(i-1, ": ", shorten_log(stat$log[[i]]), "\n"))
    }
}

---
title: Parallelize R code on a Slurm cluster
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Parallelize R code on a Slurm cluster}
  %\VignetteEngine{knitr::rmarkdown_notangle}
  %\VignetteEncoding{UTF-8}
---

Many computing-intensive processes in R involve the repeated evaluation of a
function over many items or parameter sets. These so-called [embarrassingly
parallel](https://en.wikipedia.org/wiki/Embarrassingly_parallel) calculations
can be run serially with the `lapply` or `Map` function, or in parallel on a
single machine with `mclapply` or `mcMap` (from the `parallel` package).

The rslurm package simplifies the process of distributing this type of
calculation across a computing cluster that uses the
[Slurm](http://slurm.schedmd.com/) workload manager. Its main function,
`slurm_apply` (and the related `slurm_map`) automatically divide the computation 
over multiple nodes and write the necessary submission scripts. 
The package also includes functions to retrieve
and combine the output from different nodes, as well as wrappers for common
Slurm commands.

### Table of contents

- [Basic example](#basic-example)
- [Single function evaluation](#single-function-evaluation)
- [Applying a function to a list of complex objects](#applying-a-function-to-a-list-of-complex-objects)
- [Adding auxiliary data and functions](#adding-auxiliary-data-and-functions)
- [Configuring Slurm options](#configuring-slurm-options)
- [Generating scripts for later submission](#generating-scripts-for-later-submission)
- [How it works / advanced customization](#how-it-works-advanced-customization)

## Basic example

To illustrate a typical rslurm workflow, we use a simple function that takes
a mean and standard deviation as parameters, generates a million normal deviates
and returns the sample mean and standard deviation.

```{r}
test_func <- function(par_mu, par_sd) {
    samp <- rnorm(10^6, par_mu, par_sd)
    c(s_mu = mean(samp), s_sd = sd(samp))
}
```

We then create a parameter data frame where each row is a parameter set and each
column matches an argument of the function.

```{r}
pars <- data.frame(par_mu = 1:10,
                   par_sd = seq(0.1, 1, length.out = 10))
head(pars, 3)
```

We can now pass that function and the parameters data frame to `slurm_apply`,
specifiying the number of cluster nodes to use and the number of CPUs per node.
The latter (`cpus_per_node`) determines how many processes will be forked on
each node, as the `mc.cores` argument of `parallel::mcMap`.

```{r}
library(rslurm)
sjob <- slurm_apply(test_func, pars, jobname = 'test_apply',
                    nodes = 2, cpus_per_node = 2, submit = FALSE)
```
The output of `slurm_apply` is a `slurm_job` object that stores a few pieces of 
information (job name, job ID, and the number of nodes) needed to retrieve the
job's output.

The default argument `submit = TRUE` would submit a generated script to the 
Slurm cluster and print a message confirming the job has been submitted to 
Slurm, assuming your are running R on a Slurm head node. When working from a R 
session without direct access to the cluster, you must set `submit = FALSE`. 
Either way, the function creates a folder called `\_rslurm\_[jobname]` in the 
working directory that contains scripts and data files. This folder may be moved
to a Slurm head node, the shell command `sbatch submit.sh` run from within the 
folder, and the folder moved back to your working directory. The contents of the
`\_rslurm\_[jobname]` folder after completion of the `test_apply` job, i.e.
following either manual or automatic (i.e. with `submit = TRUE`) submission to
the cluster, includes one `results_*.RDS` file for each node:

```{r}
list.files('_rslurm_test_apply', 'results')
```

The results from all the nodes can be read back into R with the
`get_slurm_out()` function. In this example, `wait = FALSE`, 
but if you use the default argument `wait = TRUE`, execution will 
be paused until the Slurm job finishes running.

```{r}
res <- get_slurm_out(sjob, outtype = 'table', wait = FALSE)
head(res, 3)
```

The utility function `print_job_status` displays the status of a submitted job
(i.e. in queue, running or completed), and `cancel_slurm` will remove a job from
the queue, aborting its execution if necessary. These functions are R wrappers
for the Slurm command line functions `squeue` and `scancel`, respectively.

When `outtype = 'table'`, the outputs from each function evaluation are 
row-bound into a single data frame; this is an appropriate format when the 
function returns a simple vector. The default `outtype = 'raw'` combines the
outputs into a list and can thus handle arbitrarily complex return objects.

```{r}
res_raw <- get_slurm_out(sjob, outtype = 'raw', wait = FALSE)
res_raw[1:3]
```

The utility function `cleanup_files` deletes the temporary folder for the
specified Slurm job.

```{r eval = FALSE}
cleanup_files(sjob)
```

## Single function evaluation

In addition to `slurm_apply`, rslurm also defines a `slurm_call` function, which
sends a single function call to the cluster. It is analogous in syntax to the 
base R function `do.call`, accepting a function and a named list of parameters
as arguments.

```{r}
sjob <- slurm_call(test_func, jobname = 'test_call',
                   list(par_mu = 5, par_sd = 1), submit = FALSE)
```

Because `slurm_call` involves a single process on a single node, it does not
recognize the `nodes` and `cpus_per_node` arguments; otherwise, it accepts the
same additional arguments (detailed in the sections below) as `slurm_apply`.

```{r eval = FALSE}
cleanup_files(sjob)
```

## Applying a function to a list of complex objects

The function passed to `slurm_apply` can only receive atomic parameters stored 
within a data frame. Suppose we want instead to apply a function `func` to a list
of complex R objects, `obj_list`. In that case we can use the function `slurm_map`,
which is similar in syntax to `lapply` from base R and `map` from the `purrr` package. 
Its first argument is a list which can contain objects of any type, and its second 
argument is a function that acts on a single element of the list.

```{r eval = FALSE}
sjob <- slurm_map(obj_list,
                  func,
                  nodes = 2, cpus_per_node = 2)
```

The output generated by `slurm_map` is structured the same way as `slurm_apply`.
The procedures for checking the job status, extracting the results of the job, and
cleaning up job files are also the same as described above.

## Adding auxiliary data and functions

Each of the tasks started by `slurm_apply` and `slurm_map` begin by default in an
"empty" R environment containing only the function to be evaluated and its parameters.
If we want to pass additional arguments to the function that do not vary with each
task, we can simply add them as additional arguments to `slurm_apply` or `slurm_map`,
like in this example, where we want to take the logarithm of many integers but always
use `log(x, base = 2)`.

```{r eval = FALSE}
sjob <- slurm_apply(log,
                    data.frame(x = 1:10000),
                    base = 2,
                    nodes = 2, cpus_per_node = 2)
```

To pass additional objects to the jobs that aren't explicitly included as arguments
to the function passed to `slurm_apply` or `slurm_map`, we can use the argument
`global_objects`. For example we might want to use an inline function that calls
two other previously defined functions.

```{r eval = FALSE}
sjob <- slurm_apply(function(a, b) c(func1(a),func2(b)), 
                    data.frame(a, b),
                    global_objects = c("func1", "func2"),
                    nodes = 2, cpus_per_node = 2)
```

The `global_objects` argument specifies the names of any R objects (besides the 
parameters data frame) that must be accessed by the function passed to 
`slurm_apply`. These objects are saved to a `.RDS` file that is loaded
on each cluster node prior to evaluating the function in parallel.

By default, all R packages attached to the current R session will also be 
attached (with `library`) on each cluster node, though this can be modified with
the optional `pkgs` argument.

## Configuring Slurm options

Particular clusters may require the specification of additional Slurm options, 
such as time and memory limits for the job. The `slurm_options` argument allows
you to set any of the command line options ([view
list](http://slurm.schedmd.com/sbatch.html)) recognized by the Slurm `sbatch`
command. It should be formatted as a named list, using the long names of each
option (e.g. "time" rather than "t"). Flags, i.e. command line options that are
toggled rather than set to a particular value, should be set to `TRUE` in
`slurm_options`. For example, the following code sets the command line options
`--time=1:00:00 --share`.

```{r eval = FALSE}
sopt <- list(time = '1:00:00', share = TRUE)
sjob <- slurm_apply(test_func, pars, slurm_options = sopt)
```

## How it works / advanced customization

As mentioned above, the `slurm_apply` function creates a job-specific folder. 
This folder contains the parameters as a RDS file and (if applicable) the
objects specified as `global_objects` saved together in a RData file. The function
also generates a R script (`slurm_run.R`) to be run on each cluster node, as
well as a Bash script (`submit.sh`) to submit the job to Slurm.

More specifically, the Bash script tells Slurm to create a job array and the R
script takes advantage of the unique `SLURM\_ARRAY\_TASK\_ID` environment
variable that Slurm will set on each cluster node. This variable is read by
`slurm_run.R`, which allows each instance of the script to operate on a
different parameter subset and write its output to a different results file. The
R script calls `parallel::mcMap` to parallelize calculations on each node.

Additionally, the `--dependency` option can be utilized by taking the job ID from the
`slurm_job` object returned by `slurm_apply`, `slurm_map`, and `slurm_call` functions.  
The ID can be manually added to the slurm options. 
In the following example, the job ID of `sjob1` is used to ensure that `sjob2` does not 
begin running until after `sjob1` finishes.

```{r eval = FALSE}
# Job1
sopt1 <- list(time = '1:00:00', share = TRUE)
sjob1 <- slurm_apply(test_func, pars, slurm_options = sopt1)

# Job2 depends on Job1
pars2 <- data.frame(par_mu = 2:20,
                   par_sd = seq(0.2, 2, length.out = 20))
sopt2 <- c(sopt1, list(dependency=sprintf("afterany:%s", sjob1$jobid)))
sjob2 <- slurm_apply(test_func2, pars2, slurm_options = sopt2)
```

Both `slurm_run.R` and `submit.sh` are generated from templates, using the 
[`whisker`](https://cran.r-project.org/package=whisker) package; these templates
can be found in the `rslurm/templates` subfolder in your R package library.
There are two templates for each script, one for `slurm_apply` and the other
(with the word "single"" in its title) for `slurm_call`.

While you should avoid changing any existing lines in the template scripts, you 
may want to add `#SBATCH` lines to the `submit.sh` templates in order to 
permanently set certain Slurm command line options and thus customize the
package to your particular cluster setup.

---
title: Parallelize R code on a Slurm cluster
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Parallelize R code on a Slurm cluster}
  %\VignetteEngine{knitr::rmarkdown_notangle}
  %\VignetteEncoding{UTF-8}
---

Many computing-intensive processes in R involve the repeated evaluation of a
function over many items or parameter sets. These so-called [embarrassingly
parallel](https://en.wikipedia.org/wiki/Embarrassingly_parallel) calculations
can be run serially with the `lapply` or `Map` function, or in parallel on a
single machine with `mclapply` or `mcMap` (from the `parallel` package).

The rslurm package simplifies the process of distributing this type of
calculation across a computing cluster that uses the
[Slurm](http://slurm.schedmd.com/) workload manager. Its main function,
`slurm_apply`, automatically divides the computation over multiple nodes and
writes the necessary submission scripts. It also includes functions to retrieve
and combine the output from different nodes, as well as wrappers for common
Slurm commands.

### Table of contents

- [Basic example](#basic-example)
- [Single function evaluation](#single-function-evaluation)
- [Adding auxiliary data and functions](#adding-auxiliary-data-and-functions)
- [Configuring Slurm options](#configuring-slurm-options)
- [Generating scripts for later submission](#generating-scripts-for-later-submission)
- [How it works / advanced customization](#how-it-works-advanced-customization)

## Basic example

To illustrate a typical rslurm workflow, we use a simple function that takes
a mean and standard deviation as parameters, generates a million normal deviates
and returns the sample mean and standard deviation.

```{r}
test_func <- function(par_mu, par_sd) {
    samp <- rnorm(10^6, par_mu, par_sd)
    c(s_mu = mean(samp), s_sd = sd(samp))
}
```

We then create a parameter data frame where each row is a parameter set and each
column matches an argument of the function.

```{r}
pars <- data.frame(par_mu = 1:10,
                   par_sd = seq(0.1, 1, length.out = 10))
head(pars, 3)
```

We can now pass that function and the parameters data frame to `slurm_apply`,
specifiying the number of cluster nodes to use and the number of CPUs per node.
The latter (`cpus_per_node`) determines how many processes will be forked on
each node, as the `mc.cores` argument of `parallel::mcMap`.

```{r}
library(rslurm)
sjob <- slurm_apply(test_func, pars, jobname = 'test_apply',
                    nodes = 2, cpus_per_node = 2, submit = FALSE)
```
The output of `slurm_apply` is a `slurm_job` object that stores a few pieces of 
information (job name, job ID, and the number of nodes) needed to retrieve the
job's output.

The default argument `submit = TRUE` would submit a generated script to the 
Slurm cluster and print a message confirming the job has been submitted to 
Slurm, assuming your are running R on a Slurm head node. When working from a R 
session without direct access to the cluster, you must set `submit = FALSE`. 
Either way, the function creates a folder called `\_rslurm\_[jobname]` in the 
working directory that contains scripts and data files. This folder may be moved
to a Slurm head node, the shell command `sbatch submit.sh` run from within the 
folder, and the folder moved back to your working directory. The contents of the
`\_rslurm\_[jobname]` folder after completion of the `test_apply` job, i.e.
following either manual or automatic (i.e. with `submit = TRUE`) submission to
the cluster, includes one `results_*.RDS` file for each node:

```{r}
list.files('_rslurm_test_apply', 'results')
```

The results from all the nodes can be read back into R with the
`get_slurm_out()` function.

```{r}
res <- get_slurm_out(sjob, outtype = 'table')
head(res, 3)
```

The utility function `print_job_status` displays the status of a submitted job
(i.e. in queue, running or completed), and `cancel_slurm` will remove a job from
the queue, aborting its execution if necessary. These functions are R wrappers
for the Slurm command line functions `squeue` and `scancel`, respectively.

When `outtype = 'table'`, the outputs from each function evaluation are 
row-bound into a single data frame; this is an appropriate format when the 
function returns a simple vector. The default `outtype = 'raw'` combines the
outputs into a list and can thus handle arbitrarily complex return objects.

```{r}
res_raw <- get_slurm_out(sjob, outtype = 'raw')
res_raw[1:3]
```

The utility function `cleanup_files` deletes the temporary folder for the
specified Slurm job.

```{r eval = FALSE}
cleanup_files(sjob)
```

## Single function evaluation

In addition to `slurm_apply`, rslurm also defines a `slurm_call` function, which
sends a single function call to the cluster. It is analogous in syntax to the 
base R function `do.call`, accepting a function and a named list of parameters
as arguments.

```{r}
sjob <- slurm_call(test_func, jobname = 'test_call',
                   list(par_mu = 5, par_sd = 1), submit = FALSE)
```

Because `slurm_call` involves a single process on a single node, it does not
recognize the `nodes` and `cpus_per_node` arguments; otherwise, it accepts the
same additional arguments (detailed in the sections below) as `slurm_apply`.

```{r eval = FALSE}
cleanup_files(sjob)
```

## Adding auxiliary data and functions

The function passed to `slurm_apply` can only receive atomic parameters stored 
within a data frame. Suppose we want instead to apply a function `func` to a list
of complex R objects, `obj_list`. To use `slurm_apply` in this case, we can wrap 
`func` in an inline function that takes an index as its sole parameter.

```{r eval = FALSE}
sjob <- slurm_apply(function(i) func(obj_list[[i]]), 
                    data.frame(i = seq_along(obj_list)),
                    add_objects = c("func", "obj_list"),
                    nodes = 2, cpus_per_node = 2)
```

The `add_objects` argument specifies the names of any R objects (besides the 
parameters data frame) that must be accessed by the function passed to 
`slurm_apply`. These objects are saved to a `.RDS` file that is loaded
on each cluster node prior to evaluating the function in parallel.

By default, all R packages attached to the current R session will also be 
attached (with `library`) on each cluster node, though this can be modified with
the optional `pkgs` argument.

## Configuring Slurm options

Particular clusters may require the specification of additional Slurm options, 
such as time and memory limits for the job. The `slurm_options` argument allows
you to set any of the command line options ([view
list](http://slurm.schedmd.com/sbatch.html)) recognized by the Slurm `sbatch`
command. It should be formatted as a named list, using the long names of each
option (e.g. "time" rather than "t"). Flags, i.e. command line options that are
toggled rather than set to a particular value, should be set to `TRUE` in
`slurm_options`. For example, the following code sets the command line options
`--time=1:00:00 --share`.

```{r eval = FALSE}
sopt <- list(time = '1:00:00', share = TRUE)
sjob <- slurm_apply(test_func, pars, slurm_options = sopt)
```

## How it works / advanced customization

As mentioned above, the `slurm_apply` function creates a job-specific folder. 
This folder contains the parameters as a RDS file and (if applicable) the
objects specified as `add_objects` saved together in a RData file. The function
also generates a R script (`slurm_run.R`) to be run on each cluster node, as
well as a Bash script (`submit.sh`) to submit the job to Slurm.

More specifically, the Bash script tells Slurm to create a job array and the R
script takes advantage of the unique `SLURM\_ARRAY\_TASK\_ID` environment
variable that Slurm will set on each cluster node. This variable is read by
`slurm_run.R`, which allows each instance of the script to operate on a
different parameter subset and write its output to a different results file. The
R script calls `parallel::mcMap` to parallelize calculations on each node.

Both `slurm_run.R` and `submit.sh` are generated from templates, using the 
[`whisker`](https://cran.r-project.org/package=whisker) package; these templates
can be found in the `rslurm/templates` subfolder in your R package library.
There are two templates for each script, one for `slurm_apply` and the other
(with the word "single"" in its title) for `slurm_call`.

While you should avoid changing any existing lines in the template scripts, you 
may want to add `#SBATCH` lines to the `submit.sh` templates in order to 
permanently set certain Slurm command line options and thus customize the
package to your particular cluster setup.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/local_slurm_array.R
\name{local_slurm_array}
\alias{local_slurm_array}
\title{Execute a Slurm job locally}
\usage{
local_slurm_array(slr_job, rscript_path = NULL)
}
\arguments{
\item{slr_job}{An object of class \code{slurm_job}.}

\item{rscript_path}{The location of the Rscript command. If not specified, 
defaults to the location of Rscript within the R installation being run.}
}
\description{
Run a previously created \code{slurm_job} object locally instead of on a
Slurm cluster
}
\details{
This function is most useful for testing your function on a reduced dataset
before submitting the full job to the Slurm cluster.

Call \code{local_slurm_array} on a \code{slurm_job} object created with
\code{slurm_apply(..., submit = FALSE)} or \code{slurm_map(..., submit = FALSE)}.
The job will run serially on the local system rather than being submitted
to the Slurm cluster.
}
\examples{
\dontrun{
sjob <- slurm_apply(func, pars, submit = FALSE)
local_slurm_array(sjob)
func_result <- get_slurm_out(sjob, "table") # Loads output data into R.
cleanup_files(sjob)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_slurm_out.R
\name{get_slurm_out}
\alias{get_slurm_out}
\title{Reads the output of a function calculated on the Slurm cluster}
\usage{
get_slurm_out(slr_job, outtype = "raw", wait = TRUE, ncores = NULL)
}
\arguments{
\item{slr_job}{A \code{slurm_job} object.}

\item{outtype}{Can be "table" or "raw", see "Value" below for details.}

\item{wait}{Specify whether to block until \code{slr_job} completes.}

\item{ncores}{(optional) If not null, the number of cores passed to
mclapply}
}
\value{
If \code{outtype = "table"}: A data frame with one column by 
  return value of the function passed to \code{slurm_apply}, where 
  each row is the output of the corresponding row in the params data frame 
  passed to \code{slurm_apply}.
  
  If \code{outtype = "raw"}: A list where each element is the output 
  of the function passed to \code{slurm_apply} for the corresponding
  row in the params data frame passed to \code{slurm_apply}.
}
\description{
This function reads all function output files (one by cluster node used) from 
the specified Slurm job and returns the result in a single data frame
(if "table" format selected) or list (if "raw" format selected). It doesn't 
record any messages (including warnings or errors) output to the R console
during the computation; these can be consulted by invoking 
\code{\link{print_job_status}}.
}
\details{
The \code{outtype} option is only relevant for jobs submitted with 
\code{slurm_apply}. Jobs sent with \code{slurm_call} only return a single
object, and setting \code{outtype = "table"} creates an error in that case.
}
\seealso{
\code{\link{slurm_apply}}, \code{\link{slurm_call}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slurm_call.R
\name{slurm_call}
\alias{slurm_call}
\title{Execution of a single function call on the Slurm cluster}
\usage{
slurm_call(
  f,
  params = list(),
  jobname = NA,
  global_objects = NULL,
  add_objects = NULL,
  pkgs = rev(.packages()),
  libPaths = NULL,
  rscript_path = NULL,
  r_template = NULL,
  sh_template = NULL,
  slurm_options = list(),
  submit = TRUE
)
}
\arguments{
\item{f}{Any R function.}

\item{params}{A named list of parameters to pass to \code{f}.}

\item{jobname}{The name of the Slurm job; if \code{NA}, it is assigned a 
random name of the form "slr####".}

\item{global_objects}{A character vector containing the name of R objects to be 
saved in a .RData file and loaded on each cluster node prior to calling 
\code{f}.}

\item{add_objects}{Older deprecated name of \code{global_objects}, retained for
backwards compatibility.}

\item{pkgs}{A character vector containing the names of packages that must be
loaded on each cluster node. By default, it includes all packages loaded by
the user when \code{slurm_call} is called.}

\item{libPaths}{A character vector describing the location of additional R
library trees to search through, or NULL. The default value of NULL
corresponds to libraries returned by \code{.libPaths()} on a cluster node.
Non-existent library trees are silently ignored.}

\item{rscript_path}{The location of the Rscript command. If not specified, 
defaults to the location of Rscript within the R installation being run.}

\item{r_template}{The path to the template file for the R script run on each node. 
If NULL, uses the default template "rslurm/templates/slurm_run_single_R.txt".}

\item{sh_template}{The path to the template file for the sbatch submission script. 
If NULL, uses the default template "rslurm/templates/submit_single_sh.txt".}

\item{slurm_options}{A named list of options recognized by \code{sbatch}; see
Details below for more information.}

\item{submit}{Whether or not to submit the job to the cluster with 
\code{sbatch}; see Details below for more information.}
}
\value{
A \code{slurm_job} object containing the \code{jobname} and the number
  of \code{nodes} effectively used.
}
\description{
Use \code{slurm_call} to perform a single function evaluation a the Slurm 
cluster.
}
\details{
This function creates a temporary folder ("_rslurm_[jobname]") in the current
directory, holding .RData and .RDS data files, the R script to run and the
Bash submission script generated for the Slurm job.

The names of any other R objects (besides \code{params}) that \code{f} needs 
to access should be listed in the \code{global_objects} argument.

Use \code{slurm_options} to set any option recognized by \code{sbatch}, e.g. 
\code{slurm_options = list(time = "1:00:00", share = TRUE)}. See
\url{http://slurm.schedmd.com/sbatch.html} for details on possible options. 
Note that full names must be used (e.g. "time" rather than "t") and that
flags (such as "share") must be specified as TRUE. The "job-name", "ntasks" 
and "output" options are already determined by \code{slurm_call} and should 
not be manually set.

When processing the computation job, the Slurm cluster will output two files 
in the temporary folder: one with the return value of the function 
("results_0.RDS") and one containing any console or error output produced by
R ("slurm_[node_id].out").

If \code{submit = TRUE}, the job is sent to the cluster and a confirmation 
message (or error) is output to the console. If \code{submit = FALSE}, a
message indicates the location of the saved data and script files; the job
can be submitted manually by running the shell command \code{sbatch
submit.sh} from that directory.

After sending the job to the Slurm cluster, \code{slurm_call} returns a 
\code{slurm_job} object which can be used to cancel the job, get the job 
status or output, and delete the temporary files associated with it. See the
description of the related functions for more details.
}
\seealso{
\code{\link{slurm_apply}} to parallelize a function over a parameter
  set.

\code{\link{cancel_slurm}}, \code{\link{cleanup_files}}, 
  \code{\link{get_slurm_out}} and \code{\link{get_job_status}} which use
  the output of this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_job_status.R
\name{get_job_status}
\alias{get_job_status}
\title{Get the status of a Slurm job}
\usage{
get_job_status(slr_job)
}
\arguments{
\item{slr_job}{A \code{slurm_job} object.}
}
\value{
A list with three elements: \code{completed} is a logical value 
  indicating if all portions of the job have completed or stopped, \code{queue} 
  contains the information on job elements still in queue, and
  \code{log} contains the console/error logs.
}
\description{
This function returns the completion status of a Slurm job, its queue status 
if any and log outputs.
}
\details{
The \code{queue} element of the output is a data frame matching the output
of the Slurm \code{squeue} command for that job; it will only indicate portions
of job that are running or in queue. The \code{log} element is a
vector of the contents of console/error output files for each node where the 
job is running.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_job_status.R, R/rslurm-deprecated.R
\name{print_job_status}
\alias{print_job_status}
\alias{rslurm-deprecated}
\title{Deprecated functions in package \pkg{rslurm}.}
\usage{
print_job_status(slr_job)
}
\description{
The functions listed below are deprecated and will be defunct in
  the near future. When possible, alternative functions with similar
  functionality are also mentioned. Help pages for deprecated functions are
  available at \code{help("-deprecated")}.
}
\section{\code{print_job_status}}{

For \code{print_job_status}, use \code{\link{get_job_status}}.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_job_status.R
\name{print_job_status-deprecated}
\alias{print_job_status-deprecated}
\title{Prints the status of a Slurm job and, if completed, its console/error output}
\usage{
print_job_status(slr_job)
}
\arguments{
\item{slr_job}{A \code{slurm_job} object.}
}
\description{
Prints the status of a Slurm job and, if completed, its console/error output.
}
\details{
If the specified Slurm job is still in the queue or running, this function
prints its current status (as output by the Slurm \code{squeue} command).
The output displays one row by node currently running part of the job ("R" in
the "ST" column) and how long it has been running ("TIME"). One row indicates
the portions of the job still in queue ("PD" in the "ST" column), if any. 

If all portions of the job have completed or stopped, the function prints the 
console and error output, if any, generated by each node.
}
\seealso{
\code{\link{rslurm-deprecated}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rslurm-package.R
\docType{package}
\name{rslurm-package}
\alias{rslurm-package}
\title{Introduction to the \code{rslurm} Package}
\description{
Send long-running or parallel jobs to a Slurm workload manager (i.e. cluster)
using the \code{\link{slurm_call}}, \code{\link{slurm_apply}}, or
\code{\link{slurm_map}} functions.
}
\section{Job submission}{

  
  This package includes three core functions used to send computations to a 
  Slurm cluster: 1) \code{\link{slurm_call}} executes a function using a 
  single set of parameters (passed as a list), 2) \code{\link{slurm_apply}} 
  evaluates a function in parallel for each row of parameters in a given 
  data frame, and 3) \code{\link{slurm_map}} evaluates a function in parallel
  for each element of a list. 
  The functions \code{slurm_apply} and \code{slurm_map} automatically split the 
  parameter rows or list elements into equal-size chunks, 
  each chunk to be processed by a separate cluster node. 
  They use functions from the \code{\link[parallel]{parallel-package}} 
  package to parallelize computations across processors on a given node.
  
  The output of \code{slurm_apply}, \code{slurm_map}, or \code{slurm_call} 
  is a \code{slurm_job} object that serves as an input to the other functions in the package: 
  \code{\link{print_job_status}}, \code{\link{cancel_slurm}}, 
  \code{\link{get_slurm_out}} and \code{\link{cleanup_files}}.
}

\section{Function specification}{

  
  To be compatible with \code{\link{slurm_apply}}, a function may accept any 
  number of single value parameters. The names of these parameters must match
  the column names of the \code{params} data frame supplied. There are no 
  restrictions on the types of parameters passed as a list to 
  \code{\link{slurm_call}} or \code{\link{slurm_map}}
  
  If the function passed to \code{slurm_call} or \code{slurm_apply} requires 
  knowledge of any R objects (data, custom helper functions) besides 
  \code{params}, a character vector corresponding to their names should be 
  passed to the optional \code{global_objects} argument.
  
  When parallelizing a function, since any error will interrupt all 
  calculations for the current node, it may be useful to wrap expressions 
  which may generate errors into a \code{\link[base]{try}} or 
  \code{\link[base:conditions]{tryCatch}} function. This will ensure the computation 
  continues with the next parameter set after reporting the error.
}

\section{Output Format}{

  
  The default output format for \code{get_slurm_out} (\code{outtype = "raw"})
  is a list where each element is the return value of one function call. If 
  the function passed to \code{slurm_apply} produces a vector output, you may
  use \code{outtype = "table"} to collect the output in a single data frame, 
  with one row by function call.
}

\section{Slurm Configuration}{

  
  Advanced options for the Slurm workload manager may accompany job submission
  by \code{\link{slurm_call}}, \code{\link{slurm_map}}, and \code{\link{slurm_apply}} 
  through the optional \code{slurm_options} argument. For example, passing
  \code{list(time = '1:30:00')} for this options limits the job to 1 hour and 30
  minutes. Some advanced configuration must be set through environment 
  variables. On a multi-cluster head node, for example, the \code{SLURM_CLUSTERS}
  environment variable must be set to direct jobs to a non-default cluster.
}

\examples{

\dontrun{
# Create a data frame of mean/sd values for normal distributions 
pars <- data.frame(par_m = seq(-10, 10, length.out = 1000), 
                   par_sd = seq(0.1, 10, length.out = 1000))
                   
# Create a function to parallelize
ftest <- function(par_m, par_sd) {
 samp <- rnorm(10^7, par_m, par_sd)
 c(s_m = mean(samp), s_sd = sd(samp))
}

sjob1 <- slurm_apply(ftest, pars)
print_job_status(sjob1)
res <- get_slurm_out(sjob1, "table")
all.equal(pars, res) # Confirm correct output
cleanup_files(sjob1)
}
  
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cancel_cleanup.R
\name{cancel_slurm}
\alias{cancel_slurm}
\title{Cancels a scheduled Slurm job}
\usage{
cancel_slurm(slr_job)
}
\arguments{
\item{slr_job}{A \code{slurm_job} object.}
}
\description{
This function cancels the specified Slurm job by invoking the Slurm 
\code{scancel} command. It does \emph{not} delete the temporary files 
(e.g. scripts) created by \code{\link{slurm_apply}} or 
\code{\link{slurm_call}}. Use \code{\link{cleanup_files}} to remove those 
files.
}
\seealso{
\code{\link{cleanup_files}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slurm_apply.R
\name{slurm_apply}
\alias{slurm_apply}
\title{Parallel execution of a function on the Slurm cluster}
\usage{
slurm_apply(
  f,
  params,
  ...,
  jobname = NA,
  nodes = 2,
  cpus_per_node = 2,
  processes_per_node = cpus_per_node,
  preschedule_cores = TRUE,
  job_array_task_limit = NULL,
  global_objects = NULL,
  add_objects = NULL,
  pkgs = rev(.packages()),
  libPaths = NULL,
  rscript_path = NULL,
  r_template = NULL,
  sh_template = NULL,
  slurm_options = list(),
  submit = TRUE
)
}
\arguments{
\item{f}{A function that accepts one or many single values as parameters and
may return any type of R object.}

\item{params}{A data frame of parameter values to apply \code{f} to. Each
column corresponds to a parameter of \code{f} (\emph{Note}: names must
match) and each row corresponds to a separate function call.}

\item{...}{Additional arguments to \code{f}. These arguments do not vary
with each call to \code{f}.}

\item{jobname}{The name of the Slurm job; if \code{NA}, it is assigned a
random name of the form "slr####".}

\item{nodes}{The (maximum) number of cluster nodes to spread the calculation
over. \code{slurm_apply} automatically divides \code{params} in chunks of
approximately equal size to send to each node. Less nodes are allocated if
the parameter set is too small to use all CPUs on the requested nodes.}

\item{cpus_per_node}{The number of CPUs requested per node. This argument is
mapped to the Slurm parameter \code{cpus-per-task}.}

\item{processes_per_node}{The number of logical CPUs to utilize per node,
i.e. how many processes to run in parallel per node. This can exceed
\code{cpus_per_node} for nodes which support hyperthreading. Defaults to
\code{processes_per_node = cpus_per_node}.}

\item{preschedule_cores}{Corresponds to the \code{mc.preschedule} argument of 
\code{parallel::mcmapply}. Defaults to \code{TRUE}. If \code{TRUE}, the 
rows of \code{params} are assigned to cores before computation. If \code{FALSE}, 
each row of \code{params} is executed by the next available core.
Setting \code{FALSE} may be faster if 
different values of \code{params} result in very variable completion time for
jobs.}

\item{job_array_task_limit}{The maximum number of job array tasks to run at 
the same time. Defaults to \code{NULL} (no limit).}

\item{global_objects}{A character vector containing the name of R objects to be
saved in a .RData file and loaded on each cluster node prior to calling
\code{f}.}

\item{add_objects}{Older deprecated name of \code{global_objects}, retained for
backwards compatibility.}

\item{pkgs}{A character vector containing the names of packages that must
be loaded on each cluster node. By default, it includes all packages
loaded by the user when \code{slurm_apply} is called.}

\item{libPaths}{A character vector describing the location of additional R
library trees to search through, or NULL. The default value of NULL
corresponds to libraries returned by \code{.libPaths()} on a cluster node.
Non-existent library trees are silently ignored.}

\item{rscript_path}{The location of the Rscript command. If not specified, 
defaults to the location of Rscript within the R installation being run.}

\item{r_template}{The path to the template file for the R script run on each node. 
If NULL, uses the default template "rslurm/templates/slurm_run_R.txt".}

\item{sh_template}{The path to the template file for the sbatch submission script. 
If NULL, uses the default template "rslurm/templates/submit_sh.txt".}

\item{slurm_options}{A named list of options recognized by \code{sbatch}; see
Details below for more information.}

\item{submit}{Whether or not to submit the job to the cluster with
\code{sbatch}; see Details below for more information.}
}
\value{
A \code{slurm_job} object containing the \code{jobname} and the
  number of \code{nodes} effectively used.
}
\description{
Use \code{slurm_apply} to compute function over multiple sets of
parameters in parallel, spread across multiple nodes of a Slurm cluster,
with similar syntax to \code{mapply}.
}
\details{
This function creates a temporary folder ("_rslurm_[jobname]") in the current
directory, holding .RData and .RDS data files, the R script to run and the Bash
submission script generated for the Slurm job.

The set of input parameters is divided in equal chunks sent to each node, and
\code{f} is evaluated in parallel within each node using functions from the
\code{parallel} R package. The names of any other R objects (besides
\code{params}) that \code{f} needs to access should be included in
\code{global_objects} or passed as additional arguments through \code{...}.

Use \code{slurm_options} to set any option recognized by \code{sbatch}, e.g.
\code{slurm_options = list(time = "1:00:00", share = TRUE)}.
See \url{http://slurm.schedmd.com/sbatch.html} for details on possible options.
Note that full names must be used (e.g. "time" rather than "t") and that flags
(such as "share") must be specified as TRUE. The "array", "job-name", "nodes", 
"cpus-per-task" and "output" options are already determined by 
\code{slurm_apply} and should not be manually set.

When processing the computation job, the Slurm cluster will output two types
of files in the temporary folder: those containing the return values of the
function for each subset of parameters ("results_[node_id].RDS") and those
containing any console or error output produced by R on each node
("slurm_[node_id].out").

If \code{submit = TRUE}, the job is sent to the cluster and a confirmation
message (or error) is output to the console. If \code{submit = FALSE},
a message indicates the location of the saved data and script files; the
job can be submitted manually by running the shell command
\code{sbatch submit.sh} from that directory.

After sending the job to the Slurm cluster, \code{slurm_apply} returns a
\code{slurm_job} object which can be used to cancel the job, get the job
status or output, and delete the temporary files associated with it. See
the description of the related functions for more details.
}
\examples{
\dontrun{
sjob <- slurm_apply(func, pars)
get_job_status(sjob) # Prints console/error output once job is completed.
func_result <- get_slurm_out(sjob, "table") # Loads output data into R.
cleanup_files(sjob)
}
}
\seealso{
\code{\link{slurm_call}} to evaluate a single function call.

\code{\link{slurm_map}} to evaluate a function over a list.

\code{\link{cancel_slurm}}, \code{\link{cleanup_files}},
  \code{\link{get_slurm_out}} and \code{\link{get_job_status}}
  which use the output of this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cancel_cleanup.R
\name{cleanup_files}
\alias{cleanup_files}
\title{Deletes temporary files associated with a Slurm job}
\usage{
cleanup_files(slr_job, wait = TRUE)
}
\arguments{
\item{slr_job}{A \code{slurm_job} object.}

\item{wait}{Specify whether to block until \code{slr_job} completes.}
}
\description{
This function deletes all temporary files associated with the specified Slurm
job, including files created by \code{\link{slurm_apply}} or 
\code{\link{slurm_call}}, as well as outputs from the cluster. These files
should be located in the \emph{_rslurm_[jobname]} folder of the current
working directory.
}
\examples{
\dontrun{
sjob <- slurm_apply(func, pars)
print_job_status(sjob) # Prints console/error output once job is completed.
func_result <- get_slurm_out(sjob, "table") # Loads output data into R.
cleanup_files(sjob)
}
}
\seealso{
\code{\link{slurm_apply}}, \code{\link{slurm_call}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slurm_job.R
\name{slurm_job}
\alias{slurm_job}
\title{Create a slurm_job object}
\usage{
slurm_job(jobname = NULL, jobid = NULL, nodes = NULL)
}
\arguments{
\item{jobname}{The name of the Slurm job. The rslurm-generated scripts and 
output files associated with a job should be found in the 
\emph{_rslurm_[jobname]} folder.}

\item{jobid}{The id of the Slurm job created by the sbatch command.}

\item{nodes}{The number of cluster nodes used by that job.}
}
\value{
A \code{slurm_job} object.
}
\description{
This function creates a \code{slurm_job} object which can be passed to other
functions such as \code{\link{cancel_slurm}}, \code{\link{cleanup_files}}, 
\code{\link{get_slurm_out}} and \code{\link{get_job_status}}.
}
\details{
In general, \code{slurm_job} objects are created automatically as the output of 
\code{\link{slurm_apply}} or \code{\link{slurm_call}}, but it may be necessary 
to manually recreate one if the job was submitted in a different R session.
}
